#!/usr/bin/env python3

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from numpy import *
from matplotlib.pyplot import *
import scipy.optimize

import unittest
import subprocess
import time
import os
import pickle
import shutil

FREECAD_BINARY = os.environ.get('TEST_FREECAD_BINARY', '/usr/bin/FreeCAD')


class TestRunNotebooks(unittest.TestCase):
  def _cleanResults(self, filename):
    baseDir = os.path.abspath(os.path.dirname(__file__))
    # remove results folder
    try:
      shutil.rmtree(baseDir+'/'+filename+'.OpticsDesign')
    except:
      pass

  def _runFile(self, filename, cleanup=True, cancelAfter=None, timeout=60*60):
    baseDir = os.path.abspath(os.path.dirname(__file__))

    # remove results folder
    if cleanup:
      try:
        shutil.rmtree(baseDir+'/'+filename+'.OpticsDesign')
      except:
        pass

    # run true simulation
    p = subprocess.Popen([FREECAD_BINARY, filename+'.FCStd', '-c'],
                          cwd=baseDir, 
                          #stdout=subprocess.DEVNULL,
                          #stderr=subprocess.DEVNULL,
                          stdin=subprocess.PIPE)

    try:
      # assemble python code to run
      pythonLines = []
      pythonLines.append('from freecad.optics_design_workbench import simulation')
      if cancelAfter is not None:
        pythonLines.append(f'import threading')
        pythonLines.append(f'import time')
        pythonLines.append(f'def delayedCancel():')
        pythonLines.append(f'  time.sleep({cancelAfter})')
        pythonLines.append(f'  simulation.runAction("stop")')
        pythonLines.append(f'')
        pythonLines.append(f'thr = threading.Thread(target=delayedCancel)')
        pythonLines.append(f'thr.start()')

      pythonLines.append('simulation.runAction("true")')

      p.stdin.write((
        '\r\n'.join(pythonLines)+'\r\nexit()\r\n'
      ).encode('utf8'))
      p.stdin.flush()

      # wait until process finishes
      t0 = time.time()
      while p.poll() is None:
        if time.time()-t0 > timeout:
          raise RuntimeError('freecad test process did not exit on time')
        time.sleep(.5)
    finally:
      p.stdin.close()
      time.sleep(1)
      if p.poll is None:
        p.kill()
        time.sleep(2)
        if p.poll is None:
          raise RuntimeError('failed killing freecad test process')

    # return paths
    return baseDir, filename


  def test_runPlaygroundExample(self):
    # make sure FCStd file runs and yields expected number of hits (> 100 rays * 10 iterations)
    baseDir, filename = self._runFile('playground')
    resultsPath = baseDir+'/'+filename+'.OpticsDesign/raw/simulation-run-000000/source-OpticalPointSource/object-OpticalAbsorberGroup'
    results = []
    for f in os.listdir(resultsPath):
      with open(resultsPath+'/'+f, 'rb') as _f:
        results.append(pickle.load(_f))
    totalHits = sum([len(r['points']) for r in results])
    self.assertGreater(totalHits, 99)
    self._cleanResults(filename)


  def _test_runWithoutSettingsObject(self):
    # make sure FCStd file without any settings can also run
    _, filename = self._runFile('nosettings')
    self._cleanResults(filename)

  def test_runAndCancelGaussianExample(self):
    t0 = time.time()
    baseDir, filename = self._runFile('gaussian', cancelAfter=5)



  def test_runGaussianExample(self):
    baseDir, filename = self._runFile('gaussian')

    # make sure results exist
    resultsPath = baseDir+'/'+filename+'.OpticsDesign/raw/simulation-run-000000/source-OpticalPointSource/object-OpticalAbsorberGroup'
    results = []
    for f in os.listdir(resultsPath):
      with open(resultsPath+'/'+f, 'rb') as _f:
        results.append(pickle.load(_f))
    totalHits = sum([len(r['points']) for r in results])
    self.assertGreater(totalHits, 1e4)

    # make sure result is perfect gaussian
    points = concatenate([r['points'] for r in results])[:,:2]
    figure()
    Hs, Xs, Ys, _ = hist2d(points[:,0], points[:,1], bins=30)
    savefig(baseDir+'/'+filename+'-gaussian-histogram.png')
    close()

    gaussian = lambda X, A, s, x0: A*exp(-(X-x0)**2/s**2)
    distance = 100
    thetaSigma = sqrt(1e-4)

    for i, (X, Y) in enumerate(
                [ ( (Xs[1:]+Xs[:-1])/2, Hs[argmin(abs(Ys)),:] ),
                  ( (Ys[1:]+Ys[:-1])/2, Hs[:,argmin(abs(Xs))] ) ]):
      figure()
      plot(X, Y, 'x')
      popt, _ = scipy.optimize.curve_fit(gaussian, X, Y, p0=(max(Y), 10, 0))
      Xlin = linspace(min(X), max(X), 300)
      Ylin = gaussian(Xlin, *popt)
      plot(Xlin, Ylin)
      plot(Xlin, gaussian(Xlin, max(Y), distance*thetaSigma, 0))
      savefig(baseDir+'/'+filename+f'-gaussian-crosssection-fit-{"xy"[i]}.pdf')
      close()
      #print(f'found sigma: {abs(popt[1]):.3f}, theoretical sigma: {distance*thetaSigma:.3f}')
      foundSigma = abs(popt[1])
      theoreticalSigma = distance*thetaSigma
      self.assertLess(abs(foundSigma-theoreticalSigma)/foundSigma, 50e-2)
      foundCenter = popt[-1]
      self.assertLess(abs(foundCenter), 50e-2)

    # cleanup
    self._cleanResults(filename)

  
  def test_runThreeTimes(self):
    # make sure working dir is clean
    baseDir, filename = self._runFile('playground')
    self._cleanResults(filename)

    # run simulation three times without cleaning
    for _ in range(3):
      baseDir, filename = self._runFile('playground', cleanup=False)
    
    # expect three result folders
    results = os.listdir(baseDir+'/'+filename+'.OpticsDesign/raw/')
    self.assertEqual(len(results), 3)

    # cleanup
    self._cleanResults(filename)


  def test_runAndCancelThreeTimes(self):
    # make sure working dir is clean
    baseDir, filename = self._runFile('gaussian', cancelAfter=3)
    self._cleanResults(filename)

    # run simulation three times without cleaning and cancel after a few seconds
    for _ in range(3):
      baseDir, filename = self._runFile('gaussian', cleanup=False, cancelAfter=3)
    
    # expect three result folders
    results = os.listdir(baseDir+'/'+filename+'.OpticsDesign/raw/')
    self.assertEqual(len(results), 3)

    # cleanup
    self._cleanResults(filename)


  def test_runFreecadNotebooks(self):
    baseDir = os.path.abspath(os.path.dirname(__file__))

    # find all notebooks and run them
    for root, dirs, files in os.walk(baseDir):
      for f in files:
        if 'notest' in f or 'notest' in root or 'checkpoint' in f or 'checkpoint' in root:
          continue
        if f.endswith('.ipynb') and not f.endswith('.nbconvert.ipynb'):
          with self.subTest(f):
            subprocess.run(f'uv run jupyter nbconvert --ExecutePreprocessor.timeout=None --to notebook --execute --inplace "{f}"',
                           cwd=root, shell=True, check=True)



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


try:
  import FreeCADGui as Gui
  import FreeCAD as App
  from FreeCAD import Vector, Rotation
  import Part
except ImportError:
  pass

from numpy import *
import scipy.optimize
import sympy as sy

from .generic_source import *
from .common import *
from . import ray
from . import find
from .. import simulation
from .. import distributions
from .. import io

#####################################################################################################
class PointSourceProxy(GenericSourceProxy):

  def onChanged(self, obj, prop):
    '''Do something when a property has changed'''

    # make sure domains are valid
    if prop in ('PhiDomain', 'ThetaDomain', 'RadiusDomain'):
      raw = getattr(obj, prop)

      # select defaults and limits depending on property
      if prop == 'PhiDomain':
        default = '0,2*pi'
        limits = ['-2*pi', '2*pi']
        spanLimits = [0, '2*pi']
      elif prop == 'ThetaDomain':
        default = '0,pi'
        limits = ['0','pi']
        spanLimits = [0, 'pi']
      elif prop == 'RadiusDomain':
        default = '0,10'
        limits = [0,inf]
        spanLimits = [0,inf]

      # parse range and replace value in case parsing changed it
      parsed, (l1,l2) = self._parsedDomain(raw, default=default, 
                                           limits=limits, 
                                           spanLimits=spanLimits)
      if raw != parsed:
        setattr(obj, prop, parsed)
      
      # inter-update theta and radius domains if possible, avoid recursion by only
      # updating the hidden one of the two domains when the non-hidden one changes
      if isfinite(float(obj.FocalLength)):
        if prop == 'ThetaDomain':
          if l1 < pi/2-1e-5 and l2 < pi/2-1e-5 and not isclose(float(obj.FocalLength), 0):
            r1, r2 = abs(tan(l1)*float(obj.FocalLength)), abs(tan(l2)*float(obj.FocalLength))
            r2 = max([0.1, r1+0.1, r2])
            setattr(obj, 'RadiusDomain', ', '.join(['0' if r==0 
                                                      else f'{r:.1e}' if abs(r)<0.01 or abs(r)>300
                                                      else f'{r:.2f}' if abs(r) < 3
                                                      else f'{r:.1f}' if abs(r) < 30
                                                      else f'{r:.0f}'
                                                          for r in sorted([r1,r2])]))
      else:
        if prop == 'RadiusDomain':
          if _lastFiniteF:=getattr(self, '_lastFiniteFocalLength', nan):
            t1, t2 = sorted([ arctan(l1/abs(_lastFiniteF)), arctan(l2/abs(_lastFiniteF)) ])
            t2 = max([0.01*pi, t1+0.01*pi, t2])
            if hasattr(obj, 'ThetaDomain'): # <- this is needed during startup to avoid 'attribute not found error'
              setattr(obj, 'ThetaDomain', ', '.join(['0' if t==0 else f'{t/pi:.2f}*pi'
                                                                    for t in [t1,t2] ]))

    # make sure resolutions are valid
    if prop in ('ThetaResolutionNumericMode', 'RadiusResolutionNumericMode', 
                'PhiResolutionNumericMode'):
      try:
        float(getattr(obj, prop))
      except Exception:
        io.err(f'tried to set {prop} to {getattr(obj, prop)}, which is not a valid number')
        if prop.startswith('Theta'):
          setattr(obj, prop, '1e6')
        else:
          setattr(obj, prop, '4')
      if float(getattr(obj, prop)) < 3:
        setattr(obj, prop, '3')
  
    # reset random number generator mode to ? if power density expression is changed
    if prop in ('PowerDensity', 'PhiDomain', 
                'ThetaDomain', 'RadiusDomain', 
                'ThetaResolutionNumericMode', 
                'RadiusResolutionNumericMode'
                'PhiResolutionNumericMode'):
      self._clearVrv(obj)

    # sync theta and radius resolution
    if prop in ('ThetaResolutionNumericMode',
                'RadiusResolutionNumericMode'):
      val = getattr(obj, prop)
      for k in ('ThetaResolutionNumericMode',
                'RadiusResolutionNumericMode'):
        # hasattr here is needed to avoid 'attribute not found' error during startup
        if k!=prop and hasattr(obj, k) and getattr(obj, k, None) != val:
          setattr(obj, prop, val)

    # parse fanPhi0
    if prop == 'FanPhi0':
      try:
        parsed = self._parsedFanPhi0(obj) 
      except Exception:
        io.err(f'invalid value for FanPhi0')
        setattr(obj, prop, '0')
      else:
        if parsed < -2*pi:
          io.err(f'value for FanPhi0 out of range [-2*pi, 2*pi]')
          setattr(obj, prop, '-2*pi')
        if parsed > 2*pi:
          io.err(f'value for FanPhi0 out of range [-2*pi, 2*pi]')
          setattr(obj, prop, '2*pi')

    # parse focal length
    if prop == 'FocalLength':
      val = getattr(obj, prop)
      try:
        val = float(val)
      except ValueError:
        setattr(obj, prop, '0')
      else:
        # show theta or radius range depending on whether focal length is finite or not
        if isfinite(float(val)):
          if float(val) != 0:
            self._lastFiniteFocalLength = float(val)
          obj.setEditorMode('ThetaDomain', 0)
          obj.setEditorMode('RadiusDomain', 3)
          obj.setEditorMode('ThetaResolutionNumericMode', 0)
          obj.setEditorMode('RadiusResolutionNumericMode', 3)
        else:
          obj.setEditorMode('ThetaDomain', 3)
          obj.setEditorMode('RadiusDomain', 0)
          obj.setEditorMode('ThetaResolutionNumericMode', 3)
          obj.setEditorMode('RadiusResolutionNumericMode', 0)

    # update divergence if focal length, emission density expression or other 
    # dependent property changed
    if prop in ('FocalLength', 'PowerDensity', 'ThetaDomain', 'ThetaResolutionNumericMode'):
      f = getattr(obj, 'FocalLength', None)
      expr = getattr(obj, 'PowerDensity', None)
      prevDivergence = getattr(obj, 'Divergence', None)
      if f is not None and expr is not None:
        f = float(f)
        if not isfinite(f):
          setattr(obj, 'Divergence', '0')
        else:
          # set divergence to read only if variables other than r exist
          # in expression
          if [str(s) for s in sy.sympify(expr).free_symbols] == ['r']:
            obj.setEditorMode('Divergence', 0)
          else:
            obj.setEditorMode('Divergence', 1)

          # create theta-only expression and find divergence angle
          expr = sy.sympify(expr).subs('r', sy.sympify(f'(tan(theta)*{f})'))
          syms = [str(s) for s in expr.free_symbols]
          if syms == ['theta'] and self.parsedThetaDomain(obj)[0] == 0:
            maxPower = sy.lambdify('theta', expr)(0)
            try:
              divergenceAngle = scipy.optimize.bisect(sy.lambdify('theta', expr - maxPower/e), 
                                                      0, self.parsedThetaDomain(obj)[1])
            except Exception:
              io.warn(f'failed to find 1/e angle in emission power expression {expr} within theta '
                      f'domain {self.parsedThetaDomain(obj)}, does the power emission decrease for '
                      f'theta>0? is the theta domain large enough?')
              setattr(obj, 'Divergence', '-')
            else:
              if (prevDivergence is None 
                    or prevDivergence == '-' 
                    or not isclose(float(eval(prevDivergence)), divergenceAngle) ):
                setattr(obj, 'Divergence', f'{-sign(f)*divergenceAngle/pi:.6g}*pi')
          else:
            setattr(obj, 'Divergence', '-')

    # update focal length if divergence changed (this can only be done if 
    # power emission is parametrized by radius only)
    if prop == 'Divergence':
      divergence = getattr(obj, 'Divergence', None)
      if divergence is not None and divergence != '-':
        newDivergenceAngle = float(eval(divergence))

      # try to find 1/e radius of power density
      f = getattr(obj, 'FocalLength', None)
      expr = getattr(obj, 'PowerDensity', None)
      if f is not None and expr is not None:
        f = float(f)
        expr = sy.sympify(expr)
        syms = [str(s) for s in expr.free_symbols]
        if syms == ['r'] and self.parsedRadiusDomain(obj)[0] == 0:
          maxPower = sy.lambdify('r', expr)(0)
          try:
            oneOverERadius = scipy.optimize.bisect(sy.lambdify('r', expr - maxPower/e), 
                                                   0, self.parsedRadiusDomain(obj)[1])
          except Exception:
            io.warn(f'failed to find 1/e angle in emission power expression {expr} within theta '
                    f'domain {self.parsedThetaDomain(obj)}, does the power emission decrease for '
                    f'theta>0? is the theta domain large enough?')
            setattr(obj, 'Divergence', '-')
          else:
            if isclose(newDivergenceAngle, 0):
              setattr(obj, 'FocalLength', 'inf')
            else:
              newFocalLength = -oneOverERadius/tan(newDivergenceAngle)
              if not isclose(newFocalLength, f, rtol=1e-5):
                setattr(obj, 'FocalLength', f'{newFocalLength:.6g}')
        else:
          setattr(obj, 'FocalLength', getattr(obj, 'FocalLength'))


  def _rvArgs(self, obj, densityString, variableDomain=None):
    useTheta = isfinite(float(obj.FocalLength))
    useRadius = not isfinite(float(obj.FocalLength))
    usePhi = variableDomain is None
    if useTheta:
      # raise error if focal length is zero and r,x or y exist in density expression (except for
      # characters in function names)
      if isclose(float(obj.FocalLength), 0):
        for c in 'rxy':
          if c in ( densityString.replace('exp', '').replace('arcsin', '').replace('arccos', '')
                                .replace('arctan', '').replace('arctan2', '').replace('arccot', '')
                                .replace('arsinh', '').replace('arcosh', '').replace('artanh', '')
                                .replace('arcoth', '') ):
            raise ValueError(f'Variable {c} in power density expression {obj.PowerDensity} '
                             f'is forbidden if focal length is zero.')

      # substitute r,x,y by theta,phi expressions
      f = f'{abs(float(obj.FocalLength)):.8e}'
      densityString = (sy.sympify(densityString)
                          .subs('r', sy.sympify(f'(tan(theta)*{f})'))
                          .subs('x', sy.sympify(f'(tan(theta)*cos(phi)*{f})'))
                          .subs('y', sy.sympify(f'(tan(theta)*sin(phi)*{f})')))
      if usePhi:
        return dict(
            probabilityDensity=densityString,
            variableOrder=('theta', 'phi'),
            variableDomains=dict(
                theta=self.parsedThetaDomain(obj), 
                phi=self.parsedPhiDomain(obj)),
            numericalResolutions=dict(
                theta=float(obj.ThetaResolutionNumericMode),
                phi=float(obj.PhiResolutionNumericMode))
        )
      else:
        return dict(
            probabilityDensity=densityString,
            variable='theta',
            variableDomain=variableDomain,
            numericalResolution=float(obj.ThetaResolutionNumericMode),
        )
    if useRadius:
      # replace sin(theta) with radius
      densityString = densityString.replace('sin(theta)', 'r')

      # raise error if theta exists in density expression
      if 'theta' in densityString:
        raise ValueError(f'Variable theta in power density expression {obj.PowerDensity} '
                         f'is forbidden if focal length is infinite.')

      # substitute theta,x,y and by r,phi expressions
      densityString = (sy.sympify(densityString)
                              .subs('x', sy.sympify(f'(r*cos(phi))'))
                              .subs('y', sy.sympify(f'(r*sin(phi))')))
      if usePhi:
        return dict(
            probabilityDensity=densityString,
            variableOrder=('r', 'phi'),
            variableDomains=dict(
                r=self.parsedRadiusDomain(obj), 
                phi=self.parsedPhiDomain(obj)),
            numericalResolutions=dict(
                r=float(obj.RadiusResolutionNumericMode),
                phi=float(obj.PhiResolutionNumericMode))
        )
      else:
        return dict(
            probabilityDensity=densityString,
            variable='r',
            variableDomain=variableDomain,
            numericalResolution=float(obj.RadiusResolutionNumericMode),
        )

  def _parsedFanPhi0(self, obj):
    return float(sy.sympify(getattr(obj, 'FanPhi0')).evalf())

  def _getVrv(self, obj):
    if NON_SERIALIZABLE_STORE.get(self, None) is None:
      NON_SERIALIZABLE_STORE[self] = {}
    
    if NON_SERIALIZABLE_STORE[self].get('vrv', None) is None:
      # attach to obj and not to self, because attributes of self should be serializable
      NON_SERIALIZABLE_STORE[self]['vrv'] = (
            distributions.VectorRandomVariable(
                **self._rvArgs(obj,
                    obj.PowerDensity+'*abs(sin(theta))', # add correction for spherical coordinate area element size 
                )
            )
      )
      vrv = NON_SERIALIZABLE_STORE[self]['vrv']
      vrv.compile()
      obj.RandomNumberGeneratorMode = vrv.mode()
    return NON_SERIALIZABLE_STORE[self]['vrv']


  def _clearVrv(self, obj):
    _stored = NON_SERIALIZABLE_STORE.get(self, {})
    _stored['vrv'] = None
    NON_SERIALIZABLE_STORE[self] = _stored
    # hasattr line is needed to avoid 'attribute not found' error during startup
    if hasattr(obj, 'RandomNumberGeneratorMode'):
      obj.RandomNumberGeneratorMode = '?'


  def _parsedDomain(self, domain, default=None, limits=None, spanLimits=None, isRecursive=False):
    # try to parse
    try:
      _domain = [float(sy.sympify(d).evalf()) for d in domain.split(',')]
    except Exception as e:
      if not isRecursive:
        io.err(f'invalid domain {domain}, {e.__class__.__name__}: {e}')
      return default, self._parsedDomain(default, None)[1]

    # make sure length is exactly two
    if _domain is not None and len(_domain) != 2:
      if not isRecursive:
        io.err(f'invalid domain {domain}, expect two numbers or inf separated by a ","')
      return default, self._parsedDomain(default, None)[1]

    # check if limits are in right order
    l1, l2 = _domain
    if l1 > l2:
      if not isRecursive:
        io.err(f'invalid domain {domain}, expect second value to be larger than first one.')
      flipped = ', '.join([s.strip() for s in reversed(domain.split(','))])
      return flipped, self._parsedDomain(flipped, None)[1]

    # check if limits are fulfilled
    if limits:
      _limits = [float(sy.sympify(l).evalf()) for l in limits]
      if l1 < _limits[0] or l2 > _limits[1]:
        if not isRecursive:
          io.err(f'domain {domain} out of bounds, expect both boundaries to be within {limits}.')
        orig1, orig2 = [s.strip() for s in domain.split(',')]
        limited = f'{limits[0] if l1 < _limits[0] else orig1}, {limits[1] if l2 > _limits[1] else orig2}'
        return limited, self._parsedDomain(limited, None)[1]

    # check if span limits are fulfilled
    if spanLimits and not isRecursive:
      _spanLimits = [float(sy.sympify(l).evalf()) for l in spanLimits]
      if l2-l1 < _spanLimits[0] or l2-l1 > _spanLimits[1]:
        # if this is a recursive call just return default to avoid possibility of endless recursion
        if isRecursive:
          return default, self._parsedDomain(default, None)[1]

        # if silence error is not set let's do our best to suggest a good domain
        else:
          io.err(f'domain span of {domain} out of bounds, expect {spanLimits[0]} <= domain span <= {spanLimits[1]} .')
          orig1, orig2 = [s.strip() for s in domain.split(',')]
          limited = f'{orig1}, {spanLimits[1] if l1==0 else {_spanLimits[1]}}'
          # silence errors and pass all limits etc. to recursive call here, because we might violate limits
          # with our enforced span limit
          return limited, self._parsedDomain(limited, default=default, limits=limits, 
                                             spanLimits=spanLimits, isRecursive=True)[1]

    # return original string and parsed domain
    return domain, _domain

  def parsedThetaDomain(self, obj):
    _, parsed = self._parsedDomain(obj.ThetaDomain)
    return parsed

  def parsedRadiusDomain(self, obj):
    _, parsed = self._parsedDomain(obj.RadiusDomain)
    return parsed

  def parsedPhiDomain(self, obj):
    _, parsed = self._parsedDomain(obj.PhiDomain)
    return parsed


  def makeRay(self, obj, theta, phi, power=1, metadata={}):
    '''
    Create new ray object with origin and direction given in global coordinates
    '''
    gpM, gpMi, opticalAxis, orthoAxis, sourceOrigin = self._makeRayCache(obj)

    # normal point source with finite focal length
    if isfinite(float(obj.FocalLength)):
      # apply azimuth and polar rotation to (0,0,1) vector
      ldirection = (Rotation(opticalAxis,phi/pi*180) 
                    * Rotation(orthoAxis,theta/pi*180) 
                    * opticalAxis)

      # shift origin to  all rays intersect in point (0,0,1)*focalLength
      lorigin = sourceOrigin + (opticalAxis-ldirection)*float(obj.FocalLength)

      # calc initial radius
      radius = tan(theta)*float(obj.FocalLength)

    # infinite focal length: passed theta is actually radius
    else:
      ldirection = opticalAxis
      lorigin = (sourceOrigin 
                  + App.Vector(theta*orthoAxis*cos(phi))
                  + App.Vector(theta*cross(orthoAxis, opticalAxis)*sin(phi)))
      radius = theta
      theta = nan
    
    # apply global placement transformation to obtain global coordinates
    p1, p2 = lorigin, lorigin+ldirection/ldirection.Length
    p1, p2 = gpM*p1, gpM*p2
    gorigin, gdirection = p1, (p2-p1)/(p2-p1).Length

    # build metadata dict
    rayMetadata = dict(initPhi=phi, initTheta=theta, initRadius=radius)
    rayMetadata.update(metadata)

    # return actual ray object
    return ray.Ray(obj, gorigin, gdirection, wavelength=obj.Wavelength, 
                   initPower=power, metadata=rayMetadata)


  def _generateRays(self, obj, mode, maxFanCount=inf, maxRaysPerFan=inf, **kwargs):
    '''
    This generator yields each ray to be traced for one simulation iteration.
    '''
    rays = []

    # make sure GUI does not freeze
    keepGuiResponsiveAndRaiseIfSimulationDone()

    # determine number of rays to place
    raysPerIteration = 100
    if settings := find.activeSimulationSettings():
      raysPerIteration = settings.RaysPerIteration
    raysPerIteration *= obj.RaysPerIterationScale

    # fan-mode: generate fans of rays in spherical coordinates
    if mode == 'fans':
      raysPerFan = min([obj.RaysPerFan, maxRaysPerFan])

      # reset flag to report fan mode once
      self._reportedFullFanMode = None

      # create obj.Fans ray fans oriented in phi0
      totalFanCount = int(min([obj.Fans, maxFanCount]))
      for fanIndex, _phi in enumerate(self._parsedFanPhi0(obj) + linspace(0, pi, totalFanCount+1)[:-1]):

        # generate the required thetas to place beams, calc the phi+pi part of the
        # fan using using positive and negative thetas here
        if isfinite(float(obj.FocalLength)):
          l1, l2 = self.parsedThetaDomain(obj)
        else:
          l1, l2 = self.parsedRadiusDomain(obj)
        if l1 == 0:
          isFullFanMode = True
          _report = getattr(self, '_reportedFullFanMode', None)
          if _report is None or _report != isFullFanMode:
            io.verb(f'using full fan-mode')
            self._reportedFullFanMode = True
          srv = distributions.ScalarRandomVariable(
                  **self._rvArgs(obj,
                      # no sin(theta) correction here because fans are 2D, but make sure theta 
                      # is taken as absolute value because of the negative-theta-hack:
                      str(sy.sympify(obj.PowerDensity).subs('theta', 'abs(theta)').subs('r', 'abs(r)')),
                      variableDomain=(-l2,l2)
                  )
          )
          # TODO: compile twice, once with _phi, once with _phi+pi, then add method to fetch the
          #       numerical densities from both compiles, merge both densities, 
          #       find grid of combined
          srv.compile(phi=_phi)
          _posNegThetas = srv.findGrid(N=raysPerFan)
          #io.verb(f'{_posNegThetas=}, {-l2=}, {l2=}')

          # store whether most central ray has positive of negative theta to
          # decide later where to place rayIndex==0
          _isCentralThetaNegative = (abs(_posNegThetas).min() < 0)

        # make one iteration per ray fan half
        for rayIndexSign, phi in ([1, _phi], [-1, _phi+pi]):

          # select thetas belonging to this half of the fan
          if l1 == 0:
            if rayIndexSign == 1:
              _thetas = _posNegThetas[_posNegThetas>=0]
            else:
              _thetas = -_posNegThetas[_posNegThetas<0]
            totalRaysInFan = len(_posNegThetas)

          # generate the required thetas to place beams in two halves if 
          # lower theta limit l1 is not zero
          else:
            isFullFanMode = False
            _report = getattr(self, '_reportedFullFanMode', None)
            if _report is None or _report != isFullFanMode:
              io.verb(f'using split fan-mode')
              self._reportedFullFanMode = False
            srv = distributions.ScalarRandomVariable(
                    **self._rvArgs(obj,
                        # no sin(theta) correction here because fans are 2D
                        obj.PowerDensity,
                        variableDomain=(l1,l2)
                    )
            )
            srv.compile(phi=_phi)
            _thetas = srv.findGrid(N=raysPerFan)
            totalRaysInFan = 2*len(_thetas)
            #io.verb(f'{_thetas=}')

          # this loop may run for quite some time, keep GUI responsive by handling events
          keepGuiResponsiveAndRaiseIfSimulationDone()

          for rayIndex, theta in enumerate(sorted(_thetas)):

            # if number of rays is even: dont use index=zero, start with +1 and
            # -1 on each side of the fan, respectively
            if raysPerFan % 2 == 0:
              rayIndex += 1

            # if number of rays is odd: place index=zero on the central theta
            else:
              if isFullFanMode:
                # increment index if we are on the side of the fan that is 
                # to avoid having two rayIndex==0 rays 
                if (       (_isCentralThetaNegative and rayIndexSign == +1)
                    or (not _isCentralThetaNegative and rayIndexSign == -1) ):
                  rayIndex += 1
              else:
                # in split ray fan mode just increment the negative ray indices
                # by one to avoid having rayIndex==0 twice
                if rayIndexSign == -1:
                  rayIndex += 1

            # this loop may run for quite some time, keep GUI responsive by handling events
            keepGuiResponsiveAndRaiseIfSimulationDone()

            # add lines corresponding to this ray to total ray list
            yield self.makeRay(obj=obj, theta=theta, phi=phi, 
                               metadata=dict(fanIndex=fanIndex, 
                                             rayIndex=rayIndex*rayIndexSign,
                                             totalFanCount=totalFanCount,
                                             totalRaysInFan=totalRaysInFan))

    # true/pseudo random mode: place rays by drawing theta and phi from true random distribution
    elif mode == 'true' or mode == 'pseudo':

      # create/get random variable for theta and phi and draw samples 
      if mode == 'true':
        thetas, phis = self._getVrv(obj).draw(N=raysPerIteration)
      elif mode == 'pseudo':
        thetas, phis = self._getVrv(obj).drawPseudo(N=raysPerIteration)

      for theta, phi in zip(thetas, phis):

        # this loop may run for quite some time, keep GUI responsive by handling events
        keepGuiResponsiveAndRaiseIfSimulationDone()

        # create and trace ray
        yield self.makeRay(obj=obj, theta=theta, phi=phi)

    else:
      raise ValueError(f'unexpected ray placement mode {mode}')


#####################################################################################################
class PointSourceViewProxy(GenericSourceViewProxy):
  pass
  
#####################################################################################################
class AddPointSource(AddGenericSource):

  def Activated(self):
    # create new feature python object
    obj = App.activeDocument().addObject('App::LinkGroupPython', 'OpticalPointSource')

    # create properties of object
    for section, entries in [
      ('OpticalEmission', [
        ('PowerDensity', 'exp(-theta^2/0.01)', 'String',  
                  'Emitted optical power per solid angle. The expression may contain any mathematical '
                  'function contained in the numpy module, the polar angle "theta" the azimuthal '
                  'angle "phi", the radial distance "r" and the cartesian coordinates "x" and "y". '
                  '"theta" and "phi" refer to a spherical coordinate system centered in the focal '
                  'point of the light source. '
                  '"r", "x" and "y" refer to a coordinate system in the emission plane (=plane '
                  'orthogonal to optical axis) centered at the intersection of emission plane and '
                  'optical axis. '
                  'All rays placed by this light source begin in the emission plane.'),
        ('Wavelength', 500, 'Float', 'Wavelength of the emitted light in nm.'),
        ('FocalLength', '0', 'String', 'Distance from the ray origin at which all rays (would) intersect. '
                  'Positive values result in converging, negative values result in a diverging '
                  'beam. For a parallel light source use "inf".'),
        ('Divergence', '-', 'String', 'Angle at which a rays at 1/e emission power diverge from the '
                  'optical axis. This option is writable only if the only unknown in the Power Density '
                  'expression is "r" (no theta, phi, x, y), if Power Density drops below 1/e for a '
                  'finite r and if Focal Length is finite. Changing divergence will update Focal Length '
                  'and vice versa. The option is readable if "theta" exists in the Power Density '
                  'expression.'),
        ('ThetaDomain', '0, pi/4', 'String', 'Min and max value for polar angle theta to consider.'),
        ('RadiusDomain', '0, 10', 'String', 'Min and max value for azimuthal angle phi to consider.'),
        ('PhiDomain', '0, 2*pi', 'String', 'Min and max value for radial distance r to consider.'),
      ]),
      ('OpticalSimulationSettings', [
        *self.defaultSimulationSettings(obj),
        ('RandomNumberGeneratorMode', '?', 'String', ''),
        ('ThetaResolutionNumericMode', '1e6', 'String', ''),
        ('RadiusResolutionNumericMode', '1e6', 'String', ''),
        ('PhiResolutionNumericMode', '3', 'String', ''),
        ('Fans', 2, 'Integer', 'Number of ray fans to place in ray fan mode.'),
        ('FanPhi0', '0', 'String', 'Change this to rotate fans around optical axis.'),
        ('RaysPerFan', 20, 'Integer', 'Number of rays to place per fan in ray fan mode.'),
       ]),
    ]:
      for name, default, kind, tooltip in entries:
        obj.addProperty('App::Property'+kind, name, section, tooltip)
        setattr(obj, name, default)

    # register custom proxy and view provider proxy
    obj.Proxy = PointSourceProxy()
    if App.GuiUp:
      obj.ViewObject.Proxy = PointSourceViewProxy()

    # make mode readonly
    obj.setEditorMode('RandomNumberGeneratorMode', 1)

    return obj

  def GetResources(self):
    return dict(Pixmap=self.iconpath(),
                Accel='',
                MenuText='Make point source',
                ToolTip='Add a point light source to the current project.')

def loadPointSource():
  Gui.addCommand('Add point source', AddPointSource())



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

try:
  import FreeCADGui as Gui
  import FreeCAD as App
  from FreeCAD import Vector, Rotation
  import Part
except ImportError:
  pass

from numpy import *
import functools

from .common import *
from . import find
from .. import simulation
from ..simulation.tracing_cache import *
from .. import io

# global dict with keys being PointSourceProxy objects and values being 
# more dicts that store pseudo-attributes. This awkward attribute storing
# format allows to bypass the serializer which wants to safe the Proxy
# objects whenever the FreeCAD project is saved.
NON_SERIALIZABLE_STORE = {}


#####################################################################################################
class GenericSourceProxy():
  '''
  Proxy of the point point source object responsible for the logic
  '''
  def execute(self, obj):
    '''Do something when doing a recomputation, this method is mandatory'''

  def onChanged(self, obj, prop):
    '''Do something when a property has changed'''
    print(obj, prop)

  def clear(self, obj):
    # delete all children, which are the RaySegments
    if hasattr(obj, 'ElementList'):
      for segment in obj.ElementList:
        simulation.simulatingDocument().removeObject(segment.Name)
    else:
      raise RuntimeError(f'light source object {obj.Name} does not have ElementList property. Older project? Recreate the source.')

  @functools.cache
  def _makeRayCache(self, obj):
    '''
    Make sure matrices and vectors that do not change during ray 
    tracing are calculated only once.
    '''
    # TODO: LinkGroupPython has no global placement method... Is this sufficient?
    gpM = obj.Placement.toMatrix()
    gpMi = obj.Placement.toMatrix().inverse()
    #elem = obj.ElementList[0]
    #gpM = obj.Shape.getGlobalPlacement().toMatrix()
    #gpMi = obj.Shape.getGlobalPlacement().toMatrix().inverse()
    
    # prepare Placement-adjusted beam orientation vectors in local coordinates
    opticalAxis = Vector(0,0,1)
    orthoAxis = Vector(1,0,0)
    sourceOrigin = Vector(0,0,0)

    return gpM, gpMi, opticalAxis, orthoAxis, sourceOrigin

  def onInitializeSimulation(self, obj, state, ident):
    pass

  def onExitSimulation(self, obj, ident):
    pass

  def runSimulationIteration(self, obj, *, mode, draw=False, store=False, **kwargs):
    # recalculate cached matrices for each iteration
    self._makeRayCache.cache_clear()

    # prepare transforms etc that will be used many times
    gpM, gpMi = self._makeRayCache(obj)[:2]

    # clear displayed rays on begin of each simulation iteration
    self.clear(obj)

    # generate rays that we want to trace in this iteration
    for ray in self._generateRays(obj, mode=mode, **kwargs):

      # add to ray object to results storage if desired
      rayResults = None
      if store and obj.RecordRays:
        if obj.RecordRays:
          rayResults = store.addRay(source=obj)

      # reference to previously drawn ray object updated in ray tracing loop, initialize
      # with ray of color given by light source
      prevRaySegment = None

      # set starting color to diffuse color of light source at begin of tracing
      # the diffuse color is the first one visible in the view settings, so it
      # is most intuitive to use this
      if _vobj:=cachedViewObject(obj):
        color = _vobj.ShapeMaterial.DiffuseColor

      # trace ray through project
      for (p1,p2), power, medium, colorChange in ray.traceRay(store=store, **kwargs):

        # this loop may run for quite some time, keep GUI responsive by handling events
        keepGuiResponsiveAndRaiseIfSimulationDone()

        # add segment to current ray in results storage if enabled
        if rayResults:
          rayResults.addSegment(points=(p1, p2), power=power, medium=medium)

        # draw line in GUI if desired
        if draw:
          # create new line element in local coordinates (global->local: transformed by inverse-GLOBAL-placement transform)
          newLineElement = Part.makeLine(gpMi*p1, gpMi*p2)

          # if color change is requested or no ray segment Part::Feature exists yet, 
          # add new Part::Feature with updated color
          if colorChange is not None or prevRaySegment is None:
            
            # calculate new color if needed
            if colorChange is not None:
              weight, newColor = colorChange
              weight = min([1, max([0, weight])])
              color = tuple(array(color)*(1-weight) + array(newColor)*weight)
          
            # create new line element and add to ray source group, set visibility to false at 
            # first to avoid rays being shown with wrong placement for a very short moment
            _obj = simulation.simulatingDocument().addObject('Part::Feature', f'RaySegment')
            _obj.Visibility = False
            if cachedViewObject(_obj):
              _obj.ViewObject.ShowInTree = False
              _obj.ViewObject.LineWidth = cachedViewObject(obj).LineWidth
              _obj.ViewObject.LineColor = color

            # make sure to create a compound with one member, instead of setting the line directly as Shape
            # setting a line directly makes the SubShapes correspond to its Vertices, which will break
            # adding possible adding of line segments to the compound in following iterations (else branch of this)
            _obj.Shape = Part.makeCompound([newLineElement])
            obj.ElementList = obj.ElementList + [_obj]
            prevRaySegment = _obj

          # if now color change is requested, add line segment as compound
          else:
            prevRaySegment.Shape = Part.makeCompound(prevRaySegment.Shape.SubShapes + [newLineElement])
        
      # increment ray count for progress tracking
      if store:
        store.incrementRayCount()

      # mark this ray is complete after tracing to permit flushing it if enabled
      if rayResults:
        rayResults.rayComplete()


#####################################################################################################
class GenericSourceViewProxy():

  def getIcon(self):
    '''Return the icon which will appear in the tree view. This method is optional and if not defined a default icon is shown.'''
    return find.iconpath(self.__class__.__name__.replace('ViewProxy', '').lower())

  def attach(self, vobj):
    NON_SERIALIZABLE_STORE[self] = vobj.Object

  def updateData(self, obj, prop):
    '''If a property of the handled feature has changed we have the chance to handle this here'''
    pass

  def claimChildren(self):
    '''Return a list of objects that will be modified by this feature'''
    return NON_SERIALIZABLE_STORE[self].ElementList

  def onDelete(self, vobj, subelements):
    # make sure all ray segments are deleted
    NON_SERIALIZABLE_STORE[self].Proxy.clear(NON_SERIALIZABLE_STORE[self])
    return True

  def onChanged(self, obj, prop):
    '''Here we can do something when a single property got changed'''
    pass


#####################################################################################################
class AddGenericSource():

  def iconpath(self):
    return find.iconpath('add-'+self.__class__.__name__.replace('Add', '').lower())

  def IsActive(self):
    return True

  def defaultSimulationSettings(self, obj):
    return [
      ('RecordRays', False, 'Bool', ''),
      ('IgnoredOpticalElements', [], 'LinkList', 'Rays of this source ignore the optical freecad_elements given'
                ' in this list.'),
      ('RaysPerIterationScale', 1, 'Float', 'Number of rays to place per simulation iteration. '
                'This will be multiplied with the RayCount property of the active simulation settings.'),
      ('MaxIntersectionsScale', 1, 'Float', 'Maximum number of intersections (reflections/refractions/'
                'detections) that ray may have with optical objects. This will be '
                'multiplied with the MaxIntersections property of the active simulation settings.'),
      ('MaxRayLengthScale', 1, 'Float', 'Maximum length of each ray segment. This will be '
                'multiplied with the MaxRayLength property of the active simulation settings.'),
    ]



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

from numpy import *
import os
import time
import pickle
import functools
import threading
import uuid
import shutil
from atomicwrites import atomic_write

from .. import freecad_elements
from .. import io
from . import processes

GET_PROGRESS_LOCK = threading.RLock()

_README_TEXT = '''
# Optics Design Workbench project folder

This folder contains simulation results, analysis and optimizer notebooks
interacting with the Optics Design Workbench. Feel free to edit this readme
document your optics design project. If this readme is deleted, it will
be restored with its default content on the next simulation run.

Subfolders raw/simulation-run-xyz will be created on each simulation run
and contain the raw ray and hit information recorded during ray tracing. 

The subfolder notebooks/ contains all jupyter notebooks used to visualize,
analyze the results and to perform sweeps and optimizations of geometry
parameters.

The optics_design_workbench.log logfile will accumulate logging messages
generated during the ray tracing for debugging purposes.
'''.strip()

def getResultsFolderPath():
  base, fname, folderName = _getFolderBase()
  return f'{base}/{folderName}'

def getLatestRunIndex():
  folderName = getResultsFolderPath()+'/raw'
  if os.path.exists(folderName):
    safeInt = lambda x: int(x) if x.isnumeric() else -1 
    return max([safeInt(f[len('simulation-run-'):])
                            for f in os.listdir(folderName) 
                                    if f.startswith('simulation-run-')]
                  +[-1])
  return -1

def generateSimulationFolderName(index=None):
  if index is None:
    index = getLatestRunIndex()+1
  return f'raw/simulation-run-{int(index):06d}'

def getLatestRunFolderPath():
  index = getLatestRunIndex()
  if index < 0:
    return None
  return f'{getResultsFolderPath()}/{generateSimulationFolderName(index)}'

def _getFolderBase():
  # check whether current file is saved
  if not processes.simulatingDocument() or not processes.simulatingDocument().getFileName():
    raise RuntimeError('cannot start simulation because no active document '
                       'or active document is not yet saved')

  # generate paths
  base, fname = os.path.split(os.path.realpath(processes.simulatingDocument().getFileName()))
  if fname.lower().endswith('.fcstd'):
    fname = fname[:-6]
  folderName = f'{fname}.OpticsDesign'
  return base, fname, folderName


class SimulationResultsSingleRay:
  def __init__(self, source):
    self.source = source
    self.isComplete = False
    self.segments = []

  def addSegment(self, points, power, medium):
    self.segments.append([points, power, medium])

  def dump(self, force=False):
    if not force and not self.isComplete:
      raise RuntimeError('trying to dump incomplete ray, this is not a good idea')

    if len(self.segments):
      # generate array of all point coordinates by taking all segment start
      # points and append the end point of the last segment
      pointArray = array([list(p) for (p,_), _, _ in self.segments]
                                                  + [self.segments[-1][0][1]])
      # generate array of all powers in segments
      powerArray = array([power for _,power,_ in self.segments])                                                  
      # generate string list of all media traversed by segments
      mediaList = [medium.Name if medium is not None else None for _,_,medium in self.segments]
      # return result dict
      return dict(points=pointArray, 
                  powers=powerArray,
                  media=mediaList)

  def rayComplete(self):
    self.isComplete = True


class SimulationResults:
  def __init__(self, simulationType, simulationRunFolder, flushEverySeconds=5, 
               dumpProgressEverySeconds=.2,
               endAfterIterations=inf, endAfterRays=inf, endAfterHits=inf):
    self.simulationType = simulationType

    self.flushEverySeconds = flushEverySeconds
    self.dumpProgressEverySeconds = dumpProgressEverySeconds

    # randomize start times to prevent synchronization of worker dumps
    self._lastFlush = time.time()+self.flushEverySeconds*random.random()
    self._lastDumpedProgress = time.time()+self.dumpProgressEverySeconds*random.random()
    self._latestProgressUpdate = time.time()
    self._lastMasterProgressDump = 0
    self._masterProgressDumpIdx = 0
    self._lastMsg = time.time()
    self.t0 = time.time()
    
    # split folder and file name and strip suffix to generate
    # basePath for this results object
    base, fname, folderName = _getFolderBase() 
    self.basePath = f'{base}/{folderName}'

    # set run folder name
    self.simulationRunFolder = simulationRunFolder

    # check whether paths are writable and mark run folder with random id to uniquely identify it in the future
    # if uid file already exists assume that this check can be skipped
    _path = f'{self.basePath}/{self.simulationRunFolder}'
    try:
      os.makedirs(_path, exist_ok=True)
      if not any([f.startswith('uid-') for f in os.listdir(_path)]):
        with open(f'{_path}/uid-{uuid.uuid4()}', 'w') as _f: 
          pass
    except Exception:
      raise RuntimeError(f'it seems simulation result path is not writable: {_path}')

    # set limitss
    self.endAfterIterations = endAfterIterations
    self.endAfterRays = endAfterRays
    self.endAfterHits = endAfterHits
    self.reachedEnd = False

    # counters for progress tracking
    self.totalIterations = 0
    self.totalTracedRays = 0
    self.totalRecordedRays = 0
    self.totalRecordedHits = 0
    self.progressByWorker = {}

    # prepare lists to store results
    self.rays = None
    self.hits = None

    # flag that is set to true when the simulation corresponding to this store is done
    self._cleanedUp = False

    # make sure the default folder structure exists, create it otherwise
    self._ensureFolderStructureExists()

  def _ensureFolderStructureExists(self):
    # create folders if not existing
    for expectPath in ['raw', 'notebooks']:
      os.makedirs(self.basePath+'/'+expectPath, exist_ok=True)

    # create readmes if not existing
    if not os.path.exists(self.basePath+'/README.md'):
      with atomic_write(self.basePath+'/README.md', mode='w') as f:
        f.write(_README_TEXT)

  def _raiseIfCleanedUp(self):
    if self._cleanedUp:
      raise RuntimeError(f'this storage was already cleaned up, cannot run requested method')

  def incrementRayCount(self):
    self.totalTracedRays += 1

  def incrementIterationCount(self):
    self.totalIterations += 1

  @functools.cache
  def _fingerprint(self):
    return f'{int(time.time()*1e3)}-pid{os.getpid()}-thread{threading.get_ident()}'

  def _makeFilename(self, kind, source=None, obj=None):
    # group data by source and by hit element
    folderName = f'{self.simulationRunFolder}'
    if type(source) is str:
      folderName += f'/{source}'
    elif source:
      folderName += f'/source-{source.Label}'
    if obj:
      folderName += f'/object-{obj.Label}'

    # generate filename from fingerprint, timestamp and kind
    fname = f'{self._fingerprint()}-{kind}.pkl'

    # make sure path exists and return
    os.makedirs(f'{self.basePath}/{folderName}', exist_ok=True)
    return f'{self.basePath}/{folderName}/{fname}'

  def flush(self):
    '''
    flush buffered data to disk and clear results lists
    '''
    # make sure instance was not yet cleaned up
    self._raiseIfCleanedUp()

    # reset fingerprint cache to make sure a fresh one is generated
    self._fingerprint.cache_clear()

    # handle ray data
    if self.rays is not None:
      # assemble dictionaries to dump
      results = {}

      # dump complete rays and store incomplete in separate list
      incomplete = []
      dump = []
      for r in self.rays:
        if r.isComplete:
          fname = self._makeFilename(kind='rays', source=r.source)
          if fname not in results.keys():
            results[fname] = []
          results[fname].append(r.dump())
        else:
          incomplete.append(r)

      # write result data
      for fname, dump in results.items():
        with open(fname, 'wb') as f:
          pickle.dump(dump, f)
        
      # replace internal ray list with incomplete list
      self.totalRecordedRays += len(self.rays or [])-len(incomplete or [])
      self.rays = incomplete if len(incomplete) else None

    # handle hit data
    if self.hits is not None:
      # assemble dictionaries to dump
      results = {}
      for source, obj, p, d, power, isEntering, metadata in self.hits:
        # create entry in results dict if needed
        fname = self._makeFilename(kind='hits', source=source, obj=obj)
        if fname not in results.keys():
          results[fname] = dict(source=source.Name, obj=obj.Name,
                                points=[], directions=[], 
                                powers=[], isEntering=[])

        # append data to entry
        results[fname]['points'].append(p)
        results[fname]['directions'].append(d)
        results[fname]['powers'].append(power)
        results[fname]['isEntering'].append(int(isEntering))

        # extend metadata entries, make sure length matches main data arrays
        currentLen = len(results[fname]['points'])-1
        for k, v in metadata.items():
          if k not in results[fname].keys():
            results[fname][k] = [nan]*currentLen
          if len(results[fname][k]) < currentLen:
            results[fname][k].extend([nan]*(currentLen-len(results[fname][k])))
          results[fname][k].append(v)

      # loop through create fnames, convert to numpy arrays and dump
      for fname, res in results.items():
        with open(fname, 'wb') as f:
          for k in res.keys():
            if type(res[k]) is list:
              # if nans exist, try to find a non-nan element. If that element has a 
              # shape, make sure nans are replaced by nan-arrays of the same shape
              if nan in res[k]:
                for _v in res[k]:
                  if not isnan(_v):
                    break
                if not isnan(_v) and len(shape(_v)):
                  for i in range(len(res[k])):
                    if isnan(res[k]):
                      res[k] = ones(shape(_v))*nan

              # try to convert to array and warn if it failed
              try:
                res[k] = array(res[k]) 
              except Exception:
                io.warn(f'failed to convert key {k} to array: {res[k]}')
          pickle.dump(res, f)

      # clear list
      self.totalRecordedHits += len(self.hits or [])
      self.hits = None

    # update last flush timestamp, add 10%ish random jitter to prevent synchronization of worker dumps
    self._lastFlush = time.time() + (.1*random.random()-.05)*self.flushEverySeconds

  def dumpProgress(self):
    '''
    dump pickled summary of simulation progress (use atomic_write)
    '''
    # make sure instance was not yet cleaned up
    self._raiseIfCleanedUp()

    # write progress atomically
    with atomic_write(self._makeFilename(source='progress', kind=str(hex(int(random.random()*1e15)))[2:]), 
                      mode='wb', overwrite=True) as f:
      pickle.dump(dict(totalIterations = self.totalIterations,
                       totalTracedRays = self.totalTracedRays,
                       totalRecordedHits = self.totalRecordedHits+len(self.hits or []),
                       totalRecordedRays = self.totalRecordedRays+len(self.rays or [])), f)

    # update last dump timestamp, add 100% random jitter to prevent synchronization of worker dumps
    self._lastDumpedProgress = time.time() + (2*random.random()-1)*self.dumpProgressEverySeconds
    self._latestProgressUpdate = time.time()

  @functools.cache
  def progressMonitorPath(self):
    return os.path.dirname(self._makeFilename(source='progress', kind='none'))

  def getProgress(self, _neverReport=False):
    # if instance is cleaned up, just return latest reported progress value
    if self._cleanedUp:
      return getattr(self, '_lastReportedProgress', {})

    # make sure no other threads interfere
    with GET_PROGRESS_LOCK:
    
      # assemble result dictionary
      result = {}
      for _, prog in self.getProgressByWorker().items():
        for k, v in prog.items():
          if k not in result:
            result[k] = 0
          result[k] += v

      # check whether simulation is done and call cancelSimulation if so
      if (result.get('totalIterations', 0) > self.endAfterIterations
              or result.get('totalTracedRays', 0) > self.endAfterRays
              or result.get('totalRecordedHits', 0) > self.endAfterHits):
        self.reachedEnd = True
        processes.setIsFinished(True)

      # dump total progress to disk from time to time
      if processes.isMasterProcess() and time.time() - self._lastMasterProgressDump > .5:
        with atomic_write(f'{self.basePath}/{self.simulationRunFolder}/progress/master-{self._masterProgressDumpIdx:09d}', 
                          mode='wb', overwrite=True) as f:
          pickle.dump(dict(totalIterations = result.get('totalIterations', 0),
                           totalTracedRays = result.get('totalTracedRays', 0),
                           totalRecordedHits = result.get('totalRecordedHits', 0),
                           totalRecordedRays = result.get('totalRecordedRays', 0),
                           endAfterIterations = self.endAfterIterations,
                           endAfterRays = self.endAfterRays,
                           endAfterHits = self.endAfterHits), f)

        # remove master progress files older than 10s
        _i = 10
        while True:
          fname = f'{self.basePath}/{self.simulationRunFolder}/progress/master-{self._masterProgressDumpIdx-_i:09d}'
          if not os.path.exists(fname):
            break
          if time.time()-os.stat(fname).st_mtime > 10:
            os.remove(fname)
          _i += 1

        # update timer and index
        self._masterProgressDumpIdx += 1
        self._lastMasterProgressDump = time.time()

      # report progress to shell from time to time
      if time.time() - self._lastMsg > 60 and not _neverReport:
        iteration = result.get("totalIterations", 0) or 0
        io.info(f'{iteration} iterations done, '
                f'{processes.isWorkerRunning()} workers are alive, '
                f'{self.performanceDescription()}')
        self._lastMsg = time.time()

      self._lastReportedProgress = result
      return result

  def performanceDescription(self):
    # disable reporting to prevent endless recursion
    p = self.getProgress(_neverReport=True)
    return (f'{60*60*p.get("totalTracedRays", 0)/(time.time()-self.t0):.1e} rays/hour, '
            f'{60*60*p.get("totalRecordedHits", 0)/(time.time()-self.t0):.1e} recorded hits/hour')

  def getProgressByWorker(self):
    # find all files in progress dir
    monitorPath = self.progressMonitorPath()
    allFiles = os.listdir(monitorPath)

    # group files by worker processes
    byWorker = {}
    for f in allFiles:
      try:
        timestamp, pid, thread, randomHash = f.split('-')
      except Exception:
        pass
      else:
        key = pid
        if key not in byWorker.keys():
          byWorker[key] = []
        byWorker[key].append([timestamp, f])

    # update state according to latest file, delete all files afterwards
    for pid, files in byWorker.items():
      if len(files):
        # load latest progress
        progress = None
        for candidate in sorted(files, key=lambda e: -int(e[0])):
          latestFile = candidate[1]
          try:
            with open(monitorPath+'/'+latestFile, 'rb') as f:
              progress = pickle.load(f)
          except Exception:
            pass
          if progress:
            break

        # update window state
        if progress and type(progress) is dict:
          if pid not in self.progressByWorker.keys():
            self.progressByWorker[pid] = {}
          self.progressByWorker[pid].update(progress)

        # delete all files
        for f in files:
          os.remove(monitorPath+'/'+f[1])

        # note current time as latest update
        self._latestProgressUpdate = time.time()
    return self.progressByWorker

  def writeDiskIfNeeded(self):
    # make sure instance was not yet cleaned up
    self._raiseIfCleanedUp()

    # check if it is time to flush to disk
    if time.time()-self._lastFlush > self.flushEverySeconds:
      self.flush()

    # check if it is time to dump progress
    if time.time()-self._lastDumpedProgress > self.dumpProgressEverySeconds:
      self.dumpProgress()

      # if we are the master process, make sure to collect all slave progresses
      # from time to time
      if processes.isMasterProcess():
        self.getProgress()

  def isSimulationRunning(self):
    return processes.isRunning()
  
  def simulationEndedGracefully(self):
    return not processes.isRunning() and self.reachedEnd

  def addRay(self, source):
    # make sure instance was not yet cleaned up
    self._raiseIfCleanedUp()

    if self.rays is None:
      self.rays = []
    ray = SimulationResultsSingleRay(source)
    self.rays.append(ray)
    self.writeDiskIfNeeded()

    # return single ray results instance
    return ray

  def addRayHit(self, source, obj, point, direction, power, isEntering, metadata):
    # make sure instance was not yet cleaned up
    self._raiseIfCleanedUp()

    if self.hits is None:
      self.hits = []
    self.hits.append([source, obj, point, direction, power, isEntering, metadata])
    self.writeDiskIfNeeded()

  def cleanup(self):
    '''
    remove temp files of the result store generated during the simulation but 
    that are not part of the actual result 
    '''
    # remove progress tracking folder
    try:
      shutil.rmtree(self.progressMonitorPath())
    except Exception:
      io.warn(f'failed to cleanup progress monitoring path at "{self.progressMonitorPath()}"')

    # mark as done
    self._cleanedUp = True



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

import os
import warnings

from . import point_source
from . import optical_group
from . import simulation_settings

def iconpath(name):
  return os.path.join(os.path.dirname(__file__), '../icons', name+'.svg')


def _allObjects():
  # this hast to be imported here to avoid circular import problems
  from .. import simulation

  for obj in simulation.simulatingDocument().Objects:
    # make sure TypeId attribute can be read without exception 
    # before yiedling to avoid yielding deleted objects
    try:
      obj.TypeId
    except Exception:
      pass
    else:
      yield obj


def lightSources():
  '''
  yield all light source objects in the current project.
  '''
  for obj in _allObjects():
    if ( obj.TypeId == 'App::LinkGroupPython'
            and isinstance(obj.Proxy, point_source.GenericSourceProxy) ):
      yield obj


def relevantOpticalObjects(lightSource=None, sequenceIndex=None):
  '''
  yield all optical objects in project
  '''
  # prepare sequential mode element lists
  sequentialModeEnable = False
  sequentialModeList = []
  currentObjectsOfSequence = []
  if _sett := activeSimulationSettings():
    sequentialModeEnable = _sett.SequentialMode
    sequentialModeList = _sett.Proxy.getTracingSequence(_sett)
  if sequenceIndex is not None and sequenceIndex < len(sequentialModeList):
    currentObjectsOfSequence = sequentialModeList[sequenceIndex]

  # prepare ignore lists
  ignoreList = []
  if lightSource:
    ignoreList = lightSource.IgnoredOpticalElements

  # loop through all objects and yield suitable objects
  for obj in _allObjects():
    if ( obj.TypeId == 'App::LinkGroupPython'
          and isinstance(obj.Proxy, optical_group.OpticalGroupProxy)
          and obj not in ignoreList
          and (not sequentialModeEnable
                or sequenceIndex is None
                or obj in currentObjectsOfSequence)):
      yield obj


def simulationSettings():
  '''
  yield all simulation object of the current project.
  '''
  for obj in _allObjects():
    if ( obj.TypeId == 'Part::FeaturePython'
            and isinstance(obj.Proxy, simulation_settings.SimulationSettingsProxy) ):
      yield obj


def activeSimulationSettings(default=None):
  '''
  Return active simulation object of the current project or None if no 
  active simulation settings object exists.
  '''
  active = []
  allSettings = []
  for obj in simulationSettings():
    allSettings.append(obj)
    if obj.Active:
      active.append(obj)
  if len(active) > 1:
    raise ValueError('only one simulation settings object may have its "Active" property '
                     'set to true, but the following objects are active: '
                     +', '.join([obj.Name for obj in active]))
  if len(active) == 1:
    return active[0]

  # if settings exist but for some reason not a single one is active, 
  # make the first one active
  if len(allSettings):
    if default is None:
      default = allSettings[0]
    warnings.warn(f'found simulation settings objects but none of them seems to '
                  f'be active, defaulting to {default.Name}')
    return default



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


import logging
import logging.handlers as handlers
import os
import datetime
import random
import warnings

_LOG_DIR                    = None
_LOGFILE_NAME               = f'optics_design_workbench.log'
_JUPYTER_REGISTERED_LOG_DIR = None 
_ROTATE_DIR                 = 'oldlogs'
_IS_INIT                    = False
_IS_VERBOSE                 = True
_MESSAGE_PREFIX             = ''

def setVerbose(isVerbose):
  global _IS_VERBOSE
  _IS_VERBOSE = bool(isVerbose)

def _logger():
  return logging.getLogger('optics_design_workbench')

def registerJupyterLogDir(path):
  global _JUPYTER_REGISTERED_LOG_DIR, _MESSAGE_PREFIX
  _JUPYTER_REGISTERED_LOG_DIR = path
  _MESSAGE_PREFIX = '(jupyter) '
  _init(forceReInit=True)

def unregisterJupyterLogDir():
  global _JUPYTER_REGISTERED_LOG_DIR, _MESSAGE_PREFIX
  _JUPYTER_REGISTERED_LOG_DIR = None
  _MESSAGE_PREFIX = ''
  _init(forceReInit=True)

def isRegisteredJupyter():
  return bool(_JUPYTER_REGISTERED_LOG_DIR)

def _getLogDir():
  if _JUPYTER_REGISTERED_LOG_DIR:
    return _JUPYTER_REGISTERED_LOG_DIR
  try:
    from .simulation import results_store
    return results_store.getResultsFolderPath()
  # runtime error is raised if no FCStd file is opened, name error is raised
  # if we are running from an freecad-external python shell, AttributeError 
  # is raised if module is not fully initialized yet. 
  # In either cases no logging is yet desired.
  except (RuntimeError, NameError, AttributeError):
    return None

def _init(forceReInit=False):
  from .simulation import processes
  global _IS_INIT, _LOG_DIR

  if forceReInit or os.path.realpath(_LOG_DIR or '') != os.path.realpath(_getLogDir() or ''):
    # isMasterProcess() returns None => situation is unclear so far, do not open log
    if processes.isMasterProcess() is None:
      return

    # isMasterProcess() returns True => we are the master process => open master log
    elif processes.isMasterProcess():
      setLogfile(_getLogDir()+'/'+_LOGFILE_NAME)

    # isMasterProcess() returns False => we are not the master process => open slave log
    else:
      setLogfile(_getLogDir()+'/'+_LOGFILE_NAME[:-4]+f'.pid{os.getpid()}')

  if _LOG_DIR is None:
    return

  os.makedirs(_LOG_DIR, exist_ok=True)
  for oldlog in [f for f in os.listdir(_LOG_DIR)
                    if f != _LOGFILE_NAME and f.startswith(_LOGFILE_NAME)]:
    os.makedirs(_LOG_DIR+'/'+_ROTATE_DIR, exist_ok=True)
    os.rename(_LOG_DIR+'/'+oldlog, _LOG_DIR+'/'+_ROTATE_DIR+'/'+oldlog)

  if not _IS_INIT:
    h = handlers.TimedRotatingFileHandler(_LOG_DIR+'/'+_LOGFILE_NAME, when='W6')
    h.setFormatter(
          logging.Formatter(
              r'%(asctime)s.%(msecs)03d000000 %(levelname)s: '
              +_MESSAGE_PREFIX
              +r'%(message)s',
              datefmt=r'%Y-%m-%dT%H:%M:%S'))
    l = _logger()
    l.addHandler(h)
    _logger().setLevel(logging.INFO)
    _IS_INIT = True

def setLogfile(name):
  '''
  change name (and path) of active logfile
  '''
  global _IS_INIT, _LOGFILE_NAME, _LOG_DIR

  # close old logfile if logger already existed
  if _IS_INIT:
    for h in _logger().handlers:
      h.close()
  name = os.path.abspath(str(name))
  baseDir = None
  if '/' in name:
    baseDir, name = os.path.split(name)
  if not name.endswith('.log'):
    name += '.log'
  _logger().handlers.clear()
  _IS_INIT = False
  _LOGFILE_NAME = name
  if baseDir is not None:
    _LOG_DIR = baseDir
  if _LOG_DIR is None:
    _LOG_DIR = '.'
  _init()

def gatherSlaveFiles():
  '''
  collect all log files of slaves and merge with master log, only runs if this
  process is the master process
  '''
  from .simulation import processes

  if not _IS_INIT or not processes.isMasterProcess():
    return

  for f in os.listdir(_LOG_DIR):
    # check if file looks like a slave's log
    if f.startswith('optics_design_workbench.pid') and f.endswith('.log'):
      pid = None
      try:
        pid = int(f[27:-4])
      except ValueError:
        pass
      if pid:

        # rename file to prevent new lines being written while we parse it
        # the slave process will recreate its own logfile if new messages appear
        while True:
          tmpName = f'{_LOG_DIR}/{int(random.random()*1e12)}.log'
          if not os.path.exists(tmpName):
            break
        os.rename(_LOG_DIR+'/'+f, tmpName)

        # append file to main log
        with open(tmpName, 'r') as inFile:
          with open(_LOG_DIR+'/'+_LOGFILE_NAME, 'a') as outFile:
            for line in inFile:
              outFile.write(f'{" ".join(line.split()[:2])} (slave {pid}) {" ".join(line.split()[2:])}\n')
        
        # remove tempfile
        os.remove(tmpName) 

def _indentMsg(msg):
  ls = [l for l in '\n'.join([str(l) for l in msg]).split('\n') if l.strip()]
  if len(ls) == 0:
    return ''
  if len(ls) == 1:
    return ls[0]
  elif len(ls) == 2:
    return ls[0]+'\n'+' '*2+r'\ '+ls[1]
  else:
    return ('\n'+' '*2+'| ').join(ls[:-1])+'\n'+' '*2+r'\ '+ls[-1]

def _prefix(kind=''):
  prefix = datetime.datetime.now().strftime('[Optics Design %H:%M:%S.%f] ')
  if kind:
    prefix += kind.upper()+': '
  return prefix

def err(*msg, logOnly=False):
  _init()
  gatherSlaveFiles()
  msg = _indentMsg(msg)
  if _logger():
    _logger().error(msg)
  if not logOnly:
    print(_prefix('error')+msg)
    if '\n' in msg:
      print()

def formatErr(*msg):
  return 'error: '+_indentMsg(msg)

def warn(*msg, logOnly=False):
  _init()
  gatherSlaveFiles()
  msg = _indentMsg(msg)
  if _logger():
    _logger().warning(msg)
  if not logOnly:
    warnings.warn(msg)
    #print(_prefix('warning')+msg)
    #if '\n' in msg:
    #  print()

def info(*msg, logOnly=None, noNewLine=False):
  # enable logOnly by default for info() and verb() in jupyter environment
  if logOnly is None:
    logOnly = isRegisteredJupyter()
  
  _init()
  gatherSlaveFiles()
  msg = _indentMsg(msg)
  if _logger():
    _logger().info(msg)
  if not logOnly:
    print(_prefix()+msg)
    if '\n' in msg and not noNewLine:
      print()

def verb(*args, **kwargs):
  if not _IS_VERBOSE:
    return
  info(*args, **kwargs)


# format times
def secondsToYMDhms(secs):
  res = []
  for mult in [365*24*60*60, 30*24*60*60, 24*60*60, 60*60, 60, 1]:
    res.append(int(secs/mult))
    secs -= res[-1]*mult
  return res

def secondsToStr(secs, length=2):
  res = []
  split = secondsToYMDhms(secs)
  for num, label in zip(split,
                        ['Y', 'M', 'd', 'h', 'm', 's']):
    if res or num:
      res.append(str(num) + label)
  return ' '.join(res[:length]) or '0s'



'''
Classes/functions handling the simulation loop, saving of results, progress tracking, background workers, etc. Submodules in here use the FreeCAD.App and Gui objects in a few places but only to fetch paths or similar metadata, not for direct interaction with FreeCAD. For direct interaction prefer to use submodules of freecad_elements.
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from .processes import *
from .results_store import *
from . import tracing_cache

def makeGoodRandomSeed():
  '''
  Make sure python and numpy random number generators both have 
  seeds that differ across threads and processes.
  '''

  import random
  import numpy.random
  import os
  import threading
  import time

  seed = int(str(threading.get_ident())+str(os.getpid())+str(int(1e7*time.time()))[-10:]) % (2**32)
  random.seed(seed)
  numpy.random.seed(seed)

# run on module load
makeGoodRandomSeed()



'''
This module contains classes/functions handling the simulation process(es).

The GUI simulations all take place in the GUI main thread using the 
"QApplication.processEvents"-hack to keep the GUI responsive. The extra
implementation work for a proper QRunnable/QThreadpool solution is 
not worth it (yet), because it would only improve the responsiveness,
not the actual simulation performance because of python's GIL.

To do simulations with actual performance gain, background processes
running headless FreeCADCmd are launched in the background.
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

from .simulation_loop import *
from .worker_process import *



'''
This module handles the simulation loop. It implements the mainloop
of the simulation and functions to start/stop/query the status for
the GUI. Only one single simulation loop can run at a time, i.e. no
fan preview is possible while another job is running.

Terminology in this module:
* master process: process that launches workers, tracks progress and 
                  cleans up in the end. The master may or may not run
                  the GUI, and it may or may not contribute to the
                  simulation work, depending how the simulation was started.
* worker process: background processes launched with subprocess module
                  that run FreeCAD in headless mode and run simulation work.
* simulation: ray-tracing procedure that accumulates a certain amount
              of rays/hits/iterations with one or more processes. A simulation
              can be canceled or finish gracefully. Status files in the 
              simulation results folder indicate that state of a simulation.
              Only one simulation for a given FCStd file can run at a time.
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

from numpy import *
import time
import datetime
import functools
import subprocess
import os
import sys
import threading
import signal
import itertools
import traceback
import tracemalloc

from ...detect_pyside import *
from ... import freecad_elements
from ... import io
from .. import results_store
from .. import tracing_cache
from . import worker_process

# fail gently if gui_windows module cannot imported
try:
  from ... import gui_windows
except Exception:
  gui_windows = None  


_TRACEMALLOC_INTERVAL = 60*60
_IS_MASTER_PROCESS = None
_SIMULATING_DOCUMENT = None
_BACKGROUND_PROCESSES = []
_ASSUME_DEAD_TIMEOUT = 15

# process info
def isMasterProcess():
  return _IS_MASTER_PROCESS

def setupJupyterMaster(path):
  '''
  Call this with a path to a FCStd file or results folder to make multiprocessing
  logic aware that this process is a jupyter process. 
  '''
  global _IS_MASTER_PROCESS
  _IS_MASTER_PROCESS = True
  # complain if already registered
  if io.isRegisteredJupyter():
    io.warn(f'setting jupyter master state even though already in jupyter master state '
            f'(this may imply a stale FreecadDocument handle is open somewhere)')
  # replace FCStd suffix
  if path.endswith('.FCStd'):
    path = path[:-6]+'.OpticsDesign'
  io.registerJupyterLogDir(path)

def jupyterBecomeSlave():
  '''
  Call this to change the role of this process to slave (e.g. when jupyter starts
  another process that will behave as a master)
  '''
  global _IS_MASTER_PROCESS
  if not io.isRegisteredJupyter():
    io.err(f'cannot use jupyterBecomeSlave if setupJupyterMaster was not called in advance')
  else:
    _IS_MASTER_PROCESS = False
    io.verb('becoming slave...')

def jupyterBecomeMaster():
  '''
  Call this to change the role of this process to master (e.g. when jupyter has
  started another process that behaved as a master, and that process finished)
  '''
  global _IS_MASTER_PROCESS
  if not io.isRegisteredJupyter():
    io.err(f'cannot use jupyterBecomeMaster if setupJupyterMaster was not called in advance')
  else:
    _IS_MASTER_PROCESS = True
    io.verb('becoming master...')

def unsetJupyterMaster():
  '''
  Call this to reset the role of this process for multiprocessing logic 
  to 'unknown'.
  '''
  global _IS_MASTER_PROCESS
  _IS_MASTER_PROCESS = None
  io.unregisterJupyterLogDir()

def isWorkerRunning():
  #print(isCanceled(), [w.isRunning() for w in _BACKGROUND_PROCESSES])
  _BACKGROUND_PROCESSES[:] = [w for w in _BACKGROUND_PROCESSES if w.isRunning()]
  return len(_BACKGROUND_PROCESSES)

def simulatingDocument():
  if _SIMULATING_DOCUMENT is not None:
    return _SIMULATING_DOCUMENT
  return App.activeDocument()

# status file info/manipulation
def _statusFilePath(name):
  return f'{results_store.getResultsFolderPath()}/{name}'

def _queryStatus(name):
  return os.path.exists(_statusFilePath(name))

def _setStatus(name, status):
  path = _statusFilePath(name)
  currentStatus = _queryStatus(name)
  if status and not currentStatus:    
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as _:
      pass
  elif not status and currentStatus:
    os.remove(path)

def isRunning():
  # if is-running file does not exist, case is closed
  if not _queryStatus('simulation-is-running'):
    return False

  # if is canceled file does not exist or a running
  # worker is known to us, assume we are still running
  if not isCanceled() or isWorkerRunning():
    return True

  # if cancel file and running file both exist and not a single worker is
  # known to this freecad process, assume run has ended without proper cleanup
  try:
    runningSince = os.stat(_statusFilePath('simulation-is-canceled')).st_mtime
    if time.time()-runningSince > _ASSUME_DEAD_TIMEOUT:
      io.warn(f'simulation was canceled {time.time()-runningSince:.1e}s ago but '
              f'is-running file still exists, assuming it died without proper '
              f'clean-up')
      setIsRunning(False)
      return False
  except Exception:
    pass

  # return true if none if the above applies  
  return True


def setIsRunning(state):
  return _setStatus('simulation-is-running', state)

def isCanceled():
  if status := _queryStatus('simulation-is-canceled'):
    try:
      setIsFinished(False)
    except Exception:
      pass
  return status

def setIsCanceled(state):
  _setStatus('simulation-is-canceled', state)

def cancelSimulation():
  if isRunning():
    setIsCanceled(True)

def isFinished():
  if status := _queryStatus('simulation-is-done'):
    try:
      setIsCanceled(False)
    except Exception:
      pass
  return status

def setIsFinished(state):
  _setStatus('simulation-is-done', state)


def runAction(action):
  '''
  This function is intended to be called by the GUI buttons and handles
  all possible actions that the buttons can trigger.
  '''
  # commands that start some sort of simulation:
  if action in ('fans', 'singlepseudo', 'singletrue', 'pseudo', 'true'):
    # simulation loop
    runSimulation(action)

  # stop button to cancel running simulation
  elif action == 'stop':
    io.info('canceling simulation...')
    cancelSimulation()


def runSimulation(action, slaveInfo={}):
  '''
  This function runs the actual work of a ray-tracing simulation and is 
  called either directly from the GUI thread or by the background worker
  processes. It keeps the GUI response using the "QApplication.processEvents"-hack
  but blocks until the simulation is done. 

  action: one of true, singletrue, pseudo, singlepseudo or fans, to determine
          how to generate rays
  slaveInfo: None if called from the master, dictionary with info passed
             from calling master process if called in a worker process
  '''
  # set global variable to mark whether we are slave or master
  global _IS_MASTER_PROCESS, _SIMULATING_DOCUMENT
  _IS_MASTER_PROCESS = not bool(slaveInfo)
  t0 = time.time()

  # reset bound box cache to prevent outdated stuff from prevailing
  tracing_cache.cacheClear()

  # setup random seeds to ensure good randomness across all workers and threads
  setupRandomSeed()

  store = None
  iteration = 0
  try:
    ##########################################################################################
    # prepare simulation, assemble simulation parameters from the various sources (settings,
    # defaults, mutual conditions, ...)

    # save active document and recompute
    _SIMULATING_DOCUMENT = App.activeDocument()
    _SIMULATING_DOCUMENT.recompute()

    # make sure other simulations have stopped and no other simulation
    # can be started
    if isMasterProcess():
      if isRunning():
        raise RuntimeError('another simulation seems to be running (or was just running and '
                           'exited ungently, in that case just retry in a few seconds)')
      setIsRunning(True)
      setIsCanceled(False)
      setIsFinished(False)
    
    # slaves expect simulation running state
    else:
      if not isRunning():
        raise RuntimeError('slave was launched but no simulation seems to be running')
        
    # determine simulation mode
    mode = action
    continuous = True
    if action.startswith('single'):
      mode = action[6:]
      continuous = False
    if action == 'fans':
      continuous = False
    
    # determine whether to store results or not
    store = False
    storeSingleShot = False
    if settings := freecad_elements.find.activeSimulationSettings():
      storeSingleShot = settings.EnableStoreSingleShotData
    if action in ('singlepseudo', 'singletrue', 'fans'):
      store = storeSingleShot
    if action in ('pseudo', 'true'):
      store = True  

    # determine whether to draw rays or not
    draw = True
    drawContinuous = False
    if settings := freecad_elements.find.activeSimulationSettings():
      drawContinuous = settings.ShowRaysInContinuousMode
    if action in ('pseudo', 'true'):
      draw = drawContinuous

    # always disable drawing in slave processes
    if not isMasterProcess():
      draw = False

    # determine number if workers to spawn (single for single shot simulations, 
    # according to settings for more than one iteration)
    workers = 1
    if continuous and (settings := freecad_elements.find.activeSimulationSettings()):
      if settings.WorkerProcessCount == 'num_cpus':
        workers = cpuCount()
      else:
        workers = int(settings.WorkerProcessCount)

    # find limits if any to stop simulation
    endAfterIterations = 10
    endAfterRays = inf
    endAfterHits = inf
    if settings := freecad_elements.find.activeSimulationSettings():
      _parse = lambda x: int(round(float(x))) if x!='inf' else inf
      endAfterIterations = _parse(settings.EndAfterIterations)
      endAfterRays = _parse(settings.EndAfterRays)
      endAfterHits = _parse(settings.EndAfterHits)

    # generate simulation run folder name
    simulationRunFolder = slaveInfo.get('simulationRunFolder', 
                                        results_store.generateSimulationFolderName())

    # generate store object and open a gui window for it (if this is not run by worker)
    if store:
      store = results_store.SimulationResults(simulationType=mode, simulationRunFolder=simulationRunFolder,
                                              endAfterIterations=endAfterIterations, 
                                              endAfterRays=endAfterRays, endAfterHits=endAfterHits)
      
      # connect progress window to this store if more than one iteration is requested
      if isMasterProcess() and continuous and gui_windows:
        gui_windows.showProgressWindow(store)

    ##########################################################################################
    # do pre-worker launched init and post-worker launched init of each light source
    # and optical object

    # run pre-worker-launch init
    if isMasterProcess():
      io.verb(f'doing pre-worker-launch init of all components...')
      for obj in itertools.chain(freecad_elements.find.lightSources(), 
                                 freecad_elements.find.relevantOpticalObjects()):
        obj.Proxy.onInitializeSimulation(obj=obj, state='pre-worker-launch', ident='master')

    # launch background worker processes (one less than specified if draw is true because 
    # then the master process will also do work), only launch if we are the master process
    if isMasterProcess():
      if draw:
        backgroundWorkers = workers-1
        io.info(f'doing simulation work with {backgroundWorkers} background workers + 1 worker running in gui process')
      else:
        backgroundWorkers = workers
        io.info(f'doing simulation work with {backgroundWorkers} background workers and lazy gui process')

      # If background workers will be started, save document so they will work on the exact
      # state of the project, but make sure to save only if GUI exists, otherwise all 
      # ViewProvider objects will break and loose their info. 
      # This implies that a master running in headless mode
      # will not save before spawning the workers (which implies nothing must be can changed 
      # in the document in headless mode)
      if backgroundWorkers > 0 and App.GuiUp:
        io.verb(f'saving document {App.GuiUp=}')
        _SIMULATING_DOCUMENT.save()
      else:
        io.verb(f'skip saving document {App.GuiUp=}')

      # actually launch workers
      for workerNo in range(backgroundWorkers):
        _BACKGROUND_PROCESSES.append(worker_process.WorkerProcess(simulationType=mode, simulationRunFolder=simulationRunFolder))
        freecad_elements.keepGuiResponsiveAndRaiseIfSimulationDone()

    # doing post-worker-launch init
    io.verb(f'doing post-worker-lauch init of all components...')
    for obj in itertools.chain(freecad_elements.find.lightSources(), 
                                freecad_elements.find.relevantOpticalObjects()):
      obj.Proxy.onInitializeSimulation(obj=obj, state='post-worker-launch', ident='master' if isMasterProcess() else 'worker')

    # report to shell that simulation starts
    if isMasterProcess():
      io.info(f'starting simulation {mode=}, {store=}, {draw=}, {workers=}, {continuous=}')

    ##########################################################################################
    # mainloop A: run actual simulation work if we are a background worker or the master
    #             with draw=True
    if not isMasterProcess() or draw:
      if isMasterProcess():
        io.verb(f'gui process is not lazy and runs the simulation mainloop')
      
      # start memory profiling
      if isfinite(_TRACEMALLOC_INTERVAL):
        tracemalloc.start()
        lastTracemallocReport = time.time()

      while True:
        # do ray-tracing for all light sources
        lightSourceExists = False
        for obj in freecad_elements.find.lightSources():
          lightSourceExists = True

          # run iteration for the light source
          obj.Proxy.runSimulationIteration(obj=obj, mode=mode, draw=draw, store=store)

          # raise simulation canceled exception if parent PID is not alive
          if not isMasterProcess():
            try:
              os.kill(slaveInfo['parentPid'], 0)
            except OSError:
              raise RuntimeError(f'parent pid {slaveInfo["parentPid"]} seems to have died, exiting as well...')

          # handle GUI events and raise if simulation is done
          freecad_elements.keepGuiResponsiveAndRaiseIfSimulationDone()

          # log top 10 biggest memory allocations
          if time.time()-lastTracemallocReport > _TRACEMALLOC_INTERVAL:
            lastTracemallocReport = time.time()
            io.verb('tracemalloc: top 10 memory allocations')
            _snapshot = tracemalloc.take_snapshot()
            _top_stats = _snapshot.statistics('lineno')
            for _stat in _top_stats[:10]:
              io.verb(f'  > {_stat}')
        
        # make sure simulation is canceled if not light source exists
        if not lightSourceExists:
          io.err(f'no light source exists in current project, cannot trace any rays.')
          raise freecad_elements.SimulationEnded()

        if store:
          # tell storage object that iteration is done
          store.incrementIterationCount()

          # makes sure disk writes are happening (without this line no progress would be written
          # to disk in the rare case of no .addRay or .addRayHit calls at all during simulation)
          store.writeDiskIfNeeded()

          if isMasterProcess():
            # make sure progress is updated in master process (this will also trigger 
            # place finished file if one of the specified end criteria is reached)
            store.getProgress()

        # end mainloop after first iteration if not in continuous (=singleshot) mode      
        if not continuous:
          raise freecad_elements.SimulationEnded()
 
      # this point should never be reached under normal conditions
      raise RuntimeError('simulation loop ended unexpectedly')

    ##########################################################################################
    # mainloop B: do not do simulation work if we are the master and draw is False, just
    #             poll progress instead, do not use a loop, but a QTimer
    io.verb(f'gui process is lazy and just tracks progress')

    timer = QTimer()
    def updateProgress():
      if store and isMasterProcess():
        # make sure progress is updated (this will also place cancel/done files if one 
        # of the specified end criteria is reached)
        store.getProgress()

      # stop if canceled or done
      if isFinished():
        io.verb('simulation is done, exiting mainloop...')
        timer.stop()

      # stop if run was canceled
      if isCanceled():
        io.info('simulation is canceled, exiting mainloop...')
        timer.stop()

    timer.timeout.connect(updateProgress)
    timer.start(300)

    # this busy-loop makes the timer useless, but it is needed because cleanup is done
    # in the finally block. Maybe restructure this in the future to improve performance
    while isWorkerRunning():
      time.sleep(1e-2)
      freecad_elements.keepGuiResponsive()

  ##########################################################################################
  # SimulationEnded exception is silently ignored
  except freecad_elements.SimulationEnded:
    pass

  # any other error cancels simulation and is re-raised
  except Exception:
    setIsCanceled(True)
    io.err(traceback.format_exc())
    raise

  # cleanup after simulation loop finishes
  finally:
    # flush store if existing
    if store and hasattr(store, 'flush'):
      store.flush()

    # worker processes just exit, master process waits until all workers
    # are finished and then sets flag files 
    if isMasterProcess():
      # set is finished flag
      if not isCanceled():
        setIsFinished(True)

      # wait for workers to finish
      _t0 = time.time()
      lastPrint = time.time()
      while isWorkerRunning():

        # keep GUI responsive and limit loop speed
        time.sleep(1e-2)
        freecad_elements.keepGuiResponsive()

        # quit/kill worker processes if they take too long
        if time.time()-_t0 > 3:
          for w in _BACKGROUND_PROCESSES:
            if time.time()-_t0 < 7:
              w.quit()
            elif time.time()-_t0 < 10:
              w.terminate()
            else:
              w.kill()

        # report progress
        if time.time()-lastPrint > 3:
          io.info(f'waiting for {len(_BACKGROUND_PROCESSES)} worker processes to finish...')
          lastPrint = time.time()

      # make sure all logfiles of worker processes are collected and merged into main log
      io.gatherSlaveFiles()

      # run simulation exit hooks
      io.verb(f'running simulation-exit hook of all components...')
      for obj in itertools.chain(freecad_elements.find.lightSources(), 
                                 freecad_elements.find.relevantOpticalObjects()):
        obj.Proxy.onExitSimulation(obj=obj, ident='master' if isMasterProcess() else 'worker')

      # reset simulating document global reference
      _SIMULATING_DOCUMENT = None

      # remove is running flag
      setIsRunning(False)

      # clean temp files if existing
      if store:
        store.cleanup()

      # report success
      performanceDescription = ''
      if store and hasattr(store, 'performanceDescription'):
        performanceDescription = f' ({store.performanceDescription()})'
      io.info(f'simulation {"ended gracefully" if not isCanceled() else "was canceled"} '
              f'after {time.time()-t0:.1e}s{performanceDescription}')


@functools.cache
def cpuCount():
  '''
  Get number of physical cpus on this machine. Tries to parse lscpu output
  and falls back to os module functions if that fails. 
  '''
  # try analyze lscpu output
  threadsPerCore = None
  coresPerSocket = None
  sockets = None
  try:
    for l in subprocess.run('lscpu', check=False, capture_output=True, text=True).stdout.split('\n'):
      if 'thread' in l.lower() and 'per core' in l.lower():
        threadsPerCore = int(l.split(':')[-1].strip())
      elif 'core' in l.lower() and 'per sock' in l.lower():
        coresPerSocket = int(l.split(':')[-1].strip())
      elif 'socket' in l.lower():
        sockets = int(l.split(':')[-1].strip())
  except Exception:
    pass

  # return number of physical cores
  if threadsPerCore and coresPerSocket and sockets:
    return coresPerSocket * sockets

  # alternatively just use standard python modules for cpu count 
  # and divide result by two because virtually all cpus use two
  # threads per core
  try:
    count = len(os.sched_getaffinity(0))
  except Exception:
    count = os.cpu_count()  
  return max([1, count//2])


def setupRandomSeed():
  # setup random seeds for numpy's and python's random module to something
  # that will differ between processes and threads for good Monte-Carlo
  # performance
  import random
  import numpy.random
  random.seed(int(abs(os.getpid()*(time.time()%1)*threading.get_ident()+1000) % 2**32))
  numpy.random.seed(int(abs(os.getpid()*(time.time()%1)*threading.get_ident()+1000) % 2**32))



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'
__version__ = None

def _determinePackageVersion():
  '''
  find out installed version of the workbench and set global variable
  '''
  global __version__
  from importlib.metadata import version
  try:
    __version__ = version('optics_design_workbench')
  except Exception:
    try:
      __version__ = version('freecad.optics_design_workbench')
    except Exception:
      __version__ = '?'

# make sure __version__ is set
_determinePackageVersion()


def _ensureSystemPackagesCanBeImported():
  '''
  make sure system packages are importable (this is important if e.g. AppImage versions are used)
  '''
  import sys
  import os
  ma, mi = sys.version_info.major, sys.version_info.minor
  found = False

  # do not add system folders if we seem to run from a venv
  if sys.prefix != sys.base_prefix:
    return

  # look for site-packages folder matching current python version, or one of the three previous
  # minor python versions or the next minor python version
  for d in (0, -1, -2, -3, 1):

    # some systems call it /usr/lib64, some call it /usr/lib, therefore check both
    for w in ('64', ''):

      # check candidate paths and add if needed
      candidate = os.path.realpath(f'/usr/lib{w}/python{ma}.{mi+d}/site-packages')
      if os.path.exists(candidate):
        if candidate not in sys.path:
          print(f'[freecad.optics_design_workbench] python package path {candidate} exists on '
                f'filesystem but not in sys.path, appending to sys.path...')
          sys.path.append(candidate)

# run on module load
_ensureSystemPackagesCanBeImported()

def versionInfo():
  '''
  print summary of version numbers that may be relevant for the workbench
  '''
  import sys
  import os
  FreeCAD = None
  freecadVersion = None
  try:
    import FreeCAD
  except ImportError:
    pass
  else:
    freecadVersion = ".".join(FreeCAD.Version()[:3])
  if freecadVersion is None:
    from . import jupyter_utils
    freecadVersion = jupyter_utils.freecadVersion()
  from . import detect_pyside
  print(f'python version:          {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}')
  print(f'workbench version:       {__version__}')
  print(f'FreeCAD version:         {freecadVersion or "?"}')
  print(f'running within FreeCAD?  {"yes" if FreeCAD else "no"}')
  if detect_pyside.detectQtMajorVersion():
    print(f'Qt major version:        {detect_pyside.detectQtMajorVersion()}')
  print(f'sys.prefix:              {sys.prefix}')
  print(f'sys.base_prefix:         {sys.base_prefix}')



'''
Each submodule here corresponds to an external window.
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from . progress import *


__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

try:
  import FreeCADGui as Gui
  import FreeCAD as App
  from FreeCAD import Vector, Rotation
  import Part
except ImportError:
  pass

from numpy import *
import tempfile
import os
import random
import functools
import pickle
import shutil

from .generic_source import *
from .common import *
from . import ray
from . import find
from .. import io
from .. import simulation


#####################################################################################################
class ReplaySourceProxy(GenericSourceProxy):

  def onChanged(self, obj, prop):
    '''Do something when a property has changed'''

  def onInitializeSimulation(self, obj, state, ident):
    # reset iterator to make all rays in stock available again
    self._yieldOriginDirectionWavelengthPower.cache_clear()

    # delete temp folder used for flag files to mark datafiles as consumed
    if state == 'pre-worker-launch' and ident == 'master':
      if os.path.exists(self.flagfiledir()):
        shutil.rmtree(self.flagfiledir())

  def onExitSimulation(self, obj, ident):
    # clear folder after simulation is done
    if ident == 'master':
      if os.path.exists(self.flagfiledir()):
        shutil.rmtree(self.flagfiledir())


  def flagfiledir(self):
    return f'{simulation.getResultsFolderPath()}/replay-source-used-files'

  def _isFileConsumed(self, obj, path):
    # find path of requested file relative to replay dir
    relpath = os.path.relpath(path, obj.ReplayFromDir)

    # map path to flagfile path, check whether it exists and create flag
    # if not the case
    flagpath = f'{self.flagfiledir()}/{relpath}'
    os.makedirs(os.path.dirname(flagpath), exist_ok=True)
    if not (consumed := os.path.exists(flagpath)):
      with open(flagpath, 'w') as _:
        pass
    return consumed


  @functools.cache
  def _yieldOriginDirectionWavelengthPower(self, obj):
    '''
    This generator yields (origin, direction, wavelength, power) tuples of all rays recorded
    on disk in randomized order. The functools.cache decorator makes sure every result is 
    only yielded once. Calling the cache_clear method resets the generator. 
    '''
    if not obj.ReplayFromDir:
      raise RuntimeError(f'please set a replay directory for light source {obj.Name} '
                         f'(Data -> Optical Emission -> Replay From Dir)')

    if not os.path.exists(obj.ReplayFromDir):
      raise RuntimeError(f'selected replay directory of light source {obj.Name} does not '
                         f'seem to exist: {obj.ReplayFromDir} ')

    io.verb(f'starting replay source iterator')
    foundHitsFile = False
    for r, ds, fs in os.walk(obj.ReplayFromDir, topdown=True):
      # go through files in random order
      random.shuffle(fs)
      random.shuffle(ds)
      for f in fs:
        if f.endswith('-hits.pkl'):
          foundHitsFile = True
          if not self._isFileConsumed(obj, f'{r}/{f}'):
            data = {}
            with open(f'{r}/{f}', 'rb') as _f:
              data = pickle.load(_f)
            indices = list(range(len(data['powers'])))
            random.shuffle(indices)
            for i in indices:
              _wavelength = 1
              if len(data.get('wavelength', [])) > i:
                _wavelength = data['wavelength'][i]
              yield (data['points'][i], data['directions'][i], 
                     _wavelength, data['powers'][i])

    # raise if not a single good datafile was found
    if not foundHitsFile:
      raise RuntimeError(f'selected replay directory of light source {obj.Name} does not '
                         f'seem to contain any ray hit datafile: {obj.ReplayFromDir} ')


  def _generateRays(self, obj, mode, **kwargs):
    '''
    This generator yields each ray to be traced for one simulation iteration.
    '''
    # make sure GUI does not freeze
    keepGuiResponsiveAndRaiseIfSimulationDone()

    # determine number of rays to place
    raysPerIteration = 100
    if settings := find.activeSimulationSettings():
      raysPerIteration = settings.RaysPerIteration
    raysPerIteration *= obj.RaysPerIterationScale

    # fan-mode: replay lightsource does not support rendering fans
    if mode == 'fans':
      io.info(f'replay light source {obj.Name} does not support ray fans and '
              f'will not place any rays.')
      return []

    # true/pseudo random mode: place rays from stored file on disk in random order until none are left
    elif mode == 'true' or mode == 'pseudo':

      # warn that pseudo random mode is not supported and behaves identical to true random
      if mode == 'pseudo':
        io.warn(f'ReplaySource does not support pseudo-random mode and behaves identical to true-random mode')

      rayCount = 0
      gpM, gpMi = self._makeRayCache(obj)[:2]

      for origin, direction, wavelength, power in self._yieldOriginDirectionWavelengthPower(obj):
        # apply placement of lightsource to coordinates
        p1, p2 = gpM*App.Vector(origin), gpM*App.Vector(origin+direction)
        gorigin = p1
        gdirection = p2-p1

        # yield created ray
        yield ray.Ray(obj, gorigin, gdirection, wavelength=wavelength, initPower=power)

        # make sure to exit if enough rays for this iteration were placed
        rayCount += 1
        if rayCount >= raysPerIteration:
          return

      io.warn(f'replay light source {obj.Name} ran out of rays, canceling simulation...')
      raise SimulationEnded()

    else:
      raise ValueError(f'unexpected ray placement mode {mode}')


#####################################################################################################
class ReplaySourceViewProxy(GenericSourceViewProxy):
  pass
  
#####################################################################################################
class AddReplaySource(AddGenericSource):

  def Activated(self):
    # create new feature python object
    obj = App.activeDocument().addObject('App::LinkGroupPython', 'OpticalReplaySource')

    # create properties of object
    for section, entries in [
      ('OpticalEmission', [
        ('ReplayFromDir', '', 'Path', 'Select a hit coordinate file generated by another '
            'simulation to replay the ray hits with this light source.'),
      ]),
      ('OpticalSimulationSettings', [
        *self.defaultSimulationSettings(obj)
      ]),
    ]:
      for name, default, kind, tooltip in entries:
        obj.addProperty('App::Property'+kind, name, section, tooltip)
        setattr(obj, name, default)

    # register custom proxy and view provider proxy
    obj.Proxy = ReplaySourceProxy()
    if App.GuiUp:
      obj.ViewObject.Proxy = ReplaySourceViewProxy()

    return obj

  def GetResources(self):
    return dict(Pixmap=self.iconpath(),
                Accel='',
                MenuText='Make replay source',
                ToolTip='Add a replay light source to the current project.')

def loadReplaySource():
  Gui.addCommand('Add replay source', AddReplaySource())



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from numpy import *
import time
import random
import functools

from . import io

class IntervalTimer:
  def __init__(self, interval, initialExecution=True):
    self.interval = interval
    if initialExecution:
      self.lastExecution = 0
    else:
      self.lastExecution = time.time()

  def check(self):
    if time.time() - self.lastExecution > self.interval:
      self.lastExecution = time.time()
      return True
    else:
      return False

  def wait(self):
    interruptableSleep(max([0, self.lastExecution + self.interval - time.time()]))
    self.lastExecution = time.time()

  def setLastExecution(self, t=None):
    if t is None:
      self.lastExecution = time.time()
    else:
      self.lastExecution = t

  def reset(self):
    self.setLastExecution()


class FrequencyTimer:
  def __init__(self, frequency, atol=0.1, rtol=1e-3):
    self.t0 = time.time()
    self.frequency = frequency
    self.lastT = time.time()
    self.atol = atol
    self.rtol = rtol
    self.asyncJumps = []

  def setFrequency(self, frequency):
    self.frequency = frequency

  def isSync(self):
    return len(asyncJumps) == 0

  def wait(self):
    # calculate target timestamp for next execution
    targetT = self.lastT + 1/self.frequency

    # if we are late by more than 1/f, shift targetT
    # to right now, store timestamp in asyncJumps
    # and return immediately
    if time.time()-targetT > 1/self.frequency:
      self.asyncJumps.append(time.time())
      self.asyncJumps = self.asyncJumps[-1000:]
      self.lastT = time.time()
      return True

    # calc absolute time tolerance from tolerances
    tol = max([self.rtol/self.frequency, self.atol, 1e-9])

    # to a coarse sleep until 100ms before targetT
    interruptableSleep(max([0, targetT-time.time() - 0.1]))

    # sleep single tolerance steps until targetT is reached
    while targetT-time.time() > tol/2:
      time.sleep(tol)
    self.lastT = targetT
    return True


class ProgressTracker:
  def __init__(self, total, pessimism=0):
    self.progress = 0
    self.total = total
    self.t0 = time.time()
    self.pessimism = pessimism

  def increment(self, increment=1):
    self.progress = min([self.progress+max([0, increment]), self.total])

  def set(self, progress, monotonic=True):
    if progress > self.progress or not monotonic:
      self.progress = min([max([0, progress]), self.total])

  def relative(self):
    if self.total != 0:
      return self.progress/self.total
    else:
      return 1

  def remaining(self):
    return io.secondsToStr(self._secondsRemaining())

  def elapsed(self):
    return io.secondsToStr(self._secondsElapsed())

  def doneTime(self):
    tend = time.time() + self._secondsRemaining()
    if tend < time.time() or tend > 1e10:
      return '--.--. --:--'
    else:
      return time.strftime('%d.%m. %H:%M', time.localtime(tend))

  def bar(self, width=100):
    if self.relative() < 1:
      bw = int(min([width-3, round((width-2)*self.relative())]))
      return '['+'='*bw+'>'+' '*int(width-3-bw)+']'
    else:
      return '['+'='*int(width-2)+']'

  def setTotal(self, total):
    self.total = total

  def _secondsElapsed(self):
    return time.time() - self.t0

  def _secondsRemaining(self):
    dt = self._secondsElapsed()
    return (dt/max([self.relative(), 1e-6]) - dt) * (1+self.pessimism)


class Condition:
  def __init__(self, conditionLambda, maxSize=1e3):
    self._lambda = conditionLambda
    self._log = []
    self._maxSize = maxSize

  def isTrue(self):
    value = self._lambda()
    self._log.append((time.time(), value))
    if len(self._log) > self._maxSize:
      self._log = self._log[::2]
    return value
  
  def isFalse(self):
    return not self.isTrue()

  def isTrueSince(self, sinceSecondsAgo):
    # record a value
    self.isTrue()

    # check log and select relevant values
    Ts, Bs = array(self._log).T
    _f = Ts>time.time()-sinceSecondsAgo

    # consider returning true only if at least one entry but not all are within time window
    if _f.any() and not _f.all():
      return Bs[_f].all()

    return False

  def isFalseSince(self, sinceSecondsAgo):
    # record a value
    self.isFalse()

    # check log and select relevant values
    Ts, Bs = array(self._log).T
    _f = Ts>time.time()-sinceSecondsAgo

    # consider returning true only if at least one entry but not all are within time window
    if _f.any() and not _f.all():
      return not Bs[_f].any()

    return False



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


from matplotlib.pyplot import *
from numpy import *

def calcHistDensity(X, bins=None):
  H, bins = histogram(X, **({} if bins is None 
                              else {'bins':bins}))
  return (bins[1:]+bins[:-1])/2, H/sum(H)
  
def calcDiffDensity(X):
  X = array(sorted(X))
  diffs = X[1:]-X[:-1]
  density = 1/(maximum(diffs, 1e-30))
  return (X[1:]+X[:-1])/2, density/sum(density)

def generatePointsWithGivenDensity1D(density, N, startFrom=None):
  # do numeric integration and normalize result to [0,1] interval
  X, Y = density
  Xi = concatenate([ [X[0]-(X[1]-X[0])/2], 
                     (X[:-1]+X[1:])/2, 
                     [X[-1]+(X[-1]-X[-2])/2] ])
  Yi = concatenate([[0],cumsum(Y)])
  Yi = (Yi-Yi.min())/(Yi.max()-Yi.min())

  # pick values from X according to inverse Yi->X mapping,
  # generate two values more than requested and strip outermost 
  # two in result to get rid of edge-errors
  Ypick = linspace(0, 1, int(round(N)))[1:-1]
  return concatenate([[X[0]], interp(Ypick, Yi, Xi), [X[-1]]])


def generatePointsWithGivenDensity2D(density, N):
  # idea: start with empty list,
  # 1) calc 2D histogram with slightly randomized bins 
  #    to avoid bin edge artifacts
  #     -> if point count and histogram deviation from target
  #        are ok, return 
  # 2) find place with highest deviation from required histogram
  # 3) add sample there
  # 4) continue from 1)
  raise RuntimeError('not implemented')



'''

'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

# import all submodules into jupyter_utils namespace
from .freecad_document import *
from .progress import *
from .hits import *
from .histogram import *
from .parameter_sweeper import *

# import a few useful toplevel modules too
from .. import timing
from .. import io



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

from . import find
from .common import *
from .. import simulation

#####################################################################################################
class SimulationSettingsProxy():
  '''
  Proxy of the point point source object responsible for the logic
  '''
  def execute(self, obj):
    '''Do something when doing a recomputation, this method is mandatory'''

  def onChanged(self, obj, prop):
    # sync Visible property with active property to allow
    # convenient spacebar toggling of active settings
    if prop == 'Visibility':
      if not obj.Visibility and obj.Active:
        obj.Visibility = True

      # only set if not equal to prevent recursion
      if obj.Active != obj.Visibility:
        obj.Active = obj.Visibility

      # copy visibility to view obj (check truth value because ViewObjects 
      # do not exist in cli mode)
      if obj.ViewObject:
        obj.ViewObject.Visibility = obj.Visibility

    if prop == 'Active':
      # only set if not equal to prevent recursion
      if obj.Visibility != obj.Active:
        obj.Visibility = obj.Active

    if prop == 'Active' and obj.Active:
      # always cancel if active settings object was changed
      for settings in find.simulationSettings():
        if settings != obj:
          settings.Active = False

    # sanitize worker count
    if prop == 'WorkerProcessCount':
      if obj.WorkerProcessCount != 'num_cpus':        
        # set to default if non-integer input was found
        count = None
        try:
          count = int(obj.WorkerProcessCount)
        except ValueError:
          obj.WorkerProcessCount = 'num_cpus'

        # limit mit to 1
        if count and count < 1:
          obj.WorkerProcessCount = str(1)
        
        # limit count to 10 times cpu count + 10 (which would never make any sense)
        if count and count > 10 + 10*simulation.cpuCount():
          obj.WorkerProcessCount = str(int(10*simulation.cpuCount()))
    
    # sanitize EndAfter settings
    if prop.startswith('EndAfter'):
      val = getattr(obj, prop)
      if val != 'inf':
        try:
          val = int(round(float(val)))
        except ValueError:
          setattr(obj, prop, 'inf')
        else:
          # limit to positive numbers
          if val < 0:
            setattr(obj, prop, '1')

    # sanitize DistanceTolerance
    if prop == 'DistanceTolerance':
      val = getattr(obj, prop)
      try:
        val = float(val)
      except ValueError:
        setattr(obj, prop, '0.01')
      else:
        # limit range
        if val < 1e-12:
          setattr(obj, prop, '1e-12')

    # if sequential mode objects are altered, make sure empties are removed
    if 'SequentialMode' in prop:
      self.getTracingSequence(obj)


  def getTracingSequence(self, obj):
    if not obj.SequentialMode:
      return []
    sequence = []
    onlyUnsetSoFar = True
    hasEmpty = False
    lastUnset = 100
    for i in reversed(range(100)):
      attr = getattr(obj, f'SequentialModeElement{i:02d}', 'unset')
      if not len(attr):
        hasEmpty = True
      else:
        # find last unset index
        if attr == 'unset' and onlyUnsetSoFar:
          lastUnset = i
        else:
          onlyUnsetSoFar = False
        # assemble sequence
        if attr and attr != 'unset':
          sequence.append(attr)
    
    # add property for last unset
    if not hasEmpty and lastUnset < 100:
      obj.addProperty('App::PropertyLinkList', f'SequentialModeElement{lastUnset:02d}', 
                      'OpticalSimulationPerformanceSettings', 
                      f'List to specify order of optical elements to '
                      f'consider during ray tracing. Rays will not collide with anything but the next object '
                      f'given in the sequence. This can cause a massive speedup. Empty list implies the non-'
                      f'sequential mode, i.e. rays can collide with any optical object in the project.')

    return list(reversed(sequence))

#####################################################################################################
class SimulationSettingsViewProxy():
  '''
  Proxy of the point point source object responsible for the view
  '''
  def __init__(self, obj):
    self.objectName = obj.Name

  def getIcon(self):
    '''Return the icon which will appear in the tree view. This method is optional and if not defined a default icon is shown.'''
    if App.activeDocument().getObject(self.objectName).Active:
      return find.iconpath('settings') 
    return find.iconpath('settings-inactive') 

  def onDelete(self, obj, subelements):
    '''Here we can do something when the feature will be deleted'''
    return True


  
#####################################################################################################
class MakeSimulationSettings:
  def Activated(self):
    # create mirror object
    obj = App.activeDocument().addObject('Part::FeaturePython', f'OpticalSimulationSettings')

    # create properties of object
    for section, entries in [
      ('OpticalSimulationSettings', [
        ('Active', True, 'Bool', 'Use these settings as simulation settings.'),
        ('EnableStoreSingleShotData', False, 'Bool', 'Store rays and hits to disk for every single-shot '
              'simulation.'),
      ]),
      ('OpticalSimulationPerformanceSettings', [
        ('EndAfterIterations', 'inf', 'String', 'Number of iterations after which simulation should stop'),
        ('EndAfterRays', '1e4', 'String', 'Number of traced rays after which simulation should stop'),
        ('EndAfterHits', 'inf', 'String', 'Number of recorded hits after which simulation should stop'),
        ('RaysPerIteration', 100, 'Float', 'Number of rays to place per simulation iteration for random '
              'and pseudo random modes.'),
        ('MaxIntersections', 100, 'Float', 'Maximum number of intersections (reflections/refractions/'
              'detections) that a ray may have with optical objects.'),
        ('DistanceTolerance', '0.01', 'String', 'If a ray is closer to a surface than this tolerance, '
              'it is considered to intersect with the surface.'),
        ('MaxRayLength', 1000, 'Float', 'Maximum length of each ray segment, i.e. the total ray length '
              'may be up to MaxIntersections*MaxRayLength. This is not a strict '
              'limit but rather a possibility for the ray tracer to save time by ignoring '
              'objects that are farther away from a given ray origin than this limit. Longer ray '
              'segments may still occur.'),
        ('ShowRaysInContinuousMode', True, 'Bool', 'Allows to switch of displaying rays in '
              'continuous simulation modes to speed up the calculation.'),
        ('WorkerProcessCount', 'num_cpus', 'String', 'Number of worker processes to spawn for continuous '
              f'simulation modes. Should be an integer or "num_cpus" ( = {simulation.cpuCount()} ).'),
        ('SequentialMode', False, 'Bool', 'Enable/disable sequential ray-tracing mode. In '
              f'sequential mode, rays will not collide with anything but the next object '
              f'given in the sequence. This can cause a massive speedup. Empty list implies the non-'
              f'sequential mode, i.e. rays can collide with any optical object in the project.'),
        ('SequentialModeElement00', None, 'LinkList', 'List to specify order of optical elements to '
              f'consider during ray tracing. Rays will not collide with anything but the next object '
              f'given in the sequence. This can cause a massive speedup. Empty list implies the non-'
              f'sequential mode, i.e. rays can collide with any optical object in the project.'),
      ]),
      ('OpticalSimulationMetadataSettings', [
        ('StoreHitInitPoint', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitInitDirection', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitInitPower', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitInitWavelength', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitInitPhi', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitInitTheta', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitRayIndex', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitFanIndex', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitTotalFanCount', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
        ('StoreHitTotalRaysInFan', False, 'Bool', 'Enable/disable additional data about light source '
              'and ray initial conditions to be stored with each ray hit.'),
      ])
    ]:
      for name, default, kind, tooltip in entries:
        obj.addProperty('App::Property'+kind, name, section, tooltip)
        setattr(obj, name, default)

    # register custom proxy and view provider proxy
    obj.Proxy = SimulationSettingsProxy()
    if App.GuiUp:
      obj.ViewObject.Proxy = SimulationSettingsViewProxy(obj)

    # set active property to true again to trigger onChange handler
    obj.Active = True

    return obj

  def IsActive(self):
    return True

  def GetResources(self):
    return dict(Pixmap=find.iconpath('settings'),
                Accel='',
                MenuText='Inserts a simulation settings object into the project tree.',
                ToolTip='Inserts a simulation settings object into the project tree. Multiple '
                        'settings objects can coexist but only one can be active at a time.')

def loadSimulationSettings():
  Gui.addCommand('Insert settings', MakeSimulationSettings())



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


#!/usr/bin/env python3

import unittest
import time
import subprocess
import os

class TestRunNotebooks(unittest.TestCase):
  def test_testRunNotebooks(self):
    baseDir = os.path.abspath(os.path.dirname(__file__))

    # find all notebooks and run them
    for root, dirs, files in os.walk(baseDir):
      for f in files:
        if f.endswith('.ipynb') and not f.endswith('.nbconvert.ipynb'):
          with self.subTest(f):
            subprocess.run(f'uv run jupyter execute "{f}"',
                           cwd=root, shell=True, check=True)



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'

try:
  import FreeCADGui as Gui
  import FreeCAD as App
except ImportError:
  pass

from . import find
from .. import simulation

# global dict with keys being PointSourceProxy objects and values being 
# more dicts that store pseudo-attributes. This awkward attribute storing
# format allows to bypass the serializer which wants to safe the Proxy
# objects whenever the FreeCAD project is saved.
NON_SERIALIZABLE_STORE = {}

#####################################################################################################
class OpticalGroupProxy():
  def execute(self, obj):
    '''Do something when doing a recomputation, this method is mandatory'''

  def setVisibleProperties(self, obj, props):
    dynamicProps = ['AbsorptionLength', 'RefractiveIndex', 'Reflectivity', 'GratingType', 
                    'GratingLinesPerMillimeter', 'GratingLinesOrientation', 
                    'GratingDiffractionOrder']
    for p in dynamicProps:
      obj.setEditorMode(p, 0 if p in props else 3)

  def onChanged(self, obj, prop):
    '''Do something when a property has changed'''

    if prop == 'OpticalType':
      oldType = getattr(self, 'oldType', None)
      newType = getattr(obj, prop)

      # set transparency to 0
      if oldType == 'Lens' and newType != 'Lens':
        for child in obj.ElementList:
          if hasattr(child.ViewObject, 'Transparency'):
            child.ViewObject.Transparency = 0
          pass

      # update which properties to display
      if newType == 'Mirror':
        self.setVisibleProperties(obj, ['Reflectivity'])
        obj.RecordHits = False

      elif newType == 'Lens':
        self.setVisibleProperties(obj, ['AbsorptionLength', 'RefractiveIndex'])
        for child in obj.ElementList:
          if hasattr(child.ViewObject, 'Transparency'):
            child.ViewObject.Transparency = 80
        obj.RecordHits = False

      elif newType == 'Grating':
        self.setVisibleProperties(obj, ['AbsorptionLength', 'RefractiveIndex',
                                  'GratingType', 'GratingLinesPerMillimeter', 
                                  'GratingLinesOrientation', 'GratingDiffractionOrder'])
        obj.RecordHits = False

      elif newType == 'Absorber':
        self.setVisibleProperties(obj, [])
        obj.RecordHits = True

      elif newType == 'Vacuum':
        self.setVisibleProperties(obj, [])
        for child in obj.ElementList:
          if hasattr(child.ViewObject, 'Transparency'):
            child.ViewObject.Transparency = 80
        obj.RecordHits = True

      # update label
      if oldType and oldType in obj.Label:
        obj.Label = obj.Label.replace(oldType, newType)

      # update old type attribute of proxy
      self.oldType = newType
    
    if prop == 'AbsorptionLength':
      val = getattr(obj, prop)
      try:
        val = float(val)
      except ValueError:
        setattr(obj, prop, 'inf')

  def onInitializeSimulation(self, obj, state, ident):
    pass

  def onExitSimulation(self, obj, ident):
    pass

  def onRayHit(self, source, obj, point, direction, power, isEntering, metadata, store):
    if store and obj.RecordHits:
      store.addRayHit(source=source, obj=obj, point=point, direction=direction, 
                      power=power, isEntering=isEntering, metadata=metadata)


#####################################################################################################
class OpticalGroupViewProxy():

  def getIcon(self):
    '''Return the icon which will appear in the tree view. This method is optional and if not defined a default icon is shown.'''
    return find.iconpath(NON_SERIALIZABLE_STORE[self].OpticalType.lower())

  def attach(self, vobj):
    NON_SERIALIZABLE_STORE[self] = vobj.Object
    pass

  def updateData(self, obj, prop):
    '''If a property of the handled feature has changed we have the chance to handle this here'''
    pass

  def onDelete(self, obj, subelements):
    '''Here we can do something when the feature will be deleted'''
    return True

  def onChanged(self, obj, prop):
    '''Here we can do something when a single property got changed'''
    pass

  
#####################################################################################################
class MakeOpticalGroup:
  def __init__(self, opticalType):
    self.opticalType = opticalType

  def Activated(self):
    # create mirror object
    obj = App.activeDocument().addObject('App::LinkGroupPython', f'Optical{self.opticalType}Group')

    # create properties of object
    for section, entries in [
      ('OpticalProperties', [
        ('OpticalType', ['Mirror', 'Lens', 'Grating', 'Absorber', 'Vacuum'], 'Enumeration', 
              'Type of the optical element, can be reflective (=Mirror), refractive (=Lens), '
              'grating (=Grating), '
              'fully absorbing (=Absorber) or completely transparent (=Vacuum).'),
        ('RefractiveIndex', 2, 'Float', 
              'Refractive index of the material used for ray tracing.'),
        ('Reflectivity', 1, 'Float', 
              'Reflectivity coefficient used for ray tracing.'),
        ('AbsorptionLength', 'inf', 'String', 
              'Optical absorption length in the material in 1/mm'),
        ('GratingType', ['Reflection', 'Transmission'], 'Enumeration', 
              'Select whether grating should be reflective of transmissive.'),
        ('GratingLinesPerMillimeter', 1000, 'Float', 
              'Number of grating lines per millimeter.'),
        ('GratingLinesOrientation', (0,0,1), 'Vector',
              'Normal of a hypothetical set of planes that intersect the grating surface, to define '
              'the rulings of the grating as these intersection lines'),
        ('GratingDiffractionOrder', 1, 'Integer', 
              'Order of diffraction at which to place refracted/transmitted rays.'),
      ]),
      ('OpticalSimulationSettings', [
        ('RecordHits', False, 'Bool', 
              'Enable recording ray hits on this optical group to disk during simulations.'),
      ]),
    ]:
      for name, default, kind, tooltip in entries:
        obj.addProperty('App::Property'+kind, name, section, tooltip)
        setattr(obj, name, default)

    # create custom view object properties
    for section, entries in [
      ('ColorizeRays', [
        ('Weight', 0, 'Float', 
              'Weight of ray colorization, should be between 0 and 1. 1 means set color '
              'immediately, 0 means do not change color at all.'),
        ('Color', (0.,0.,0.,0.), 'Color', 
              'Color to mix with previous ray color during colorization. Weight determines '
              'fraction of old and new colors in mix.'),
      ]),
    ]:
      for name, default, kind, tooltip in entries:
        obj.ViewObject.addProperty('App::Property'+kind, name, section, tooltip)
        setattr(obj.ViewObject, name, default)

    # register custom proxy and view provider proxy
    obj.Proxy = OpticalGroupProxy()
    if App.GuiUp:
      obj.ViewObject.Proxy = OpticalGroupViewProxy()

    # set OpticalType property again to trigger onChange handler
    obj.OpticalType = self.opticalType

    # add selection to group
    obj.ElementList = Gui.Selection.getSelection()

    return obj

  def IsActive(self):
    return True

  def GetResources(self):
    return dict(Pixmap=find.iconpath('add-'+self.opticalType.lower()),
                Accel='',
                MenuText='Make '+dict(
                            Mirror='mirrors',
                            Lens='lenses',
                            Grating='gratings',
                            Absorber='absorbers',
                            Vacuum='detectors')[self.opticalType],
                ToolTip='Turn selected objects into optical '+dict(
                            Mirror='mirrors',
                            Lens='lenses',
                            Grating='gratings',
                            Absorber='absorbers',
                            Vacuum='detectors')[self.opticalType],)

def loadGroups():
  Gui.addCommand('Make mirror', MakeOpticalGroup('Mirror'))
  Gui.addCommand('Make lens', MakeOpticalGroup('Lens'))
  Gui.addCommand('Make grating', MakeOpticalGroup('Grating'))
  Gui.addCommand('Make absorber', MakeOpticalGroup('Absorber'))
  Gui.addCommand('Make detector', MakeOpticalGroup('Vacuum'))



__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


try:
  import FreeCADGui as Gui
  import FreeCAD as App
  from FreeCAD import Vector, Rotation
  import Part
except ImportError:
  pass

from numpy import *
import time

from . import find
from ..simulation.tracing_cache import *
from .common import *


class Ray():
  '''
  Class representing an individual ray.
  '''
  def __init__(self, lightSource, initPoint, initDirection, 
               wavelength, initPower=1, metadata={}):
    self.lightSource = lightSource
    self.initPoint = initPoint
    self.initDirection = initDirection
    self.initPower = initPower
    self.initWavelength = wavelength
    self.metadata = metadata
  
     
  def traceRay(self, powerTol=1e-6, maxRayLength=None,
               maxIntersections=None, store=False,
               metadata={}):
    '''
    Find all reflection/refraction/detection points of this ray. Returns a
    generator that yields (p1,p2), power, medium tuples. p1, p2 are two
    vectors describing a ray segment. power is the ray power at p1. medium
    is None for vacuum or the FreeCAD object if the ray is traveling through
    and optical object.
    '''
    if _sett := find.activeSimulationSettings():
      # calc ray length and intersection limits from settings
      if maxRayLength is None:
        maxRayLength = ( self.lightSource.MaxRayLengthScale 
                          * _sett.MaxRayLength )
      if maxIntersections is None:
        maxIntersections = (self.lightSource.MaxIntersectionsScale
                              *_sett.MaxIntersections )

      # construct ray metadata dictionary according to selected properties
      # in simulation settings
      _enabledKeys = [k.lower()[8:] for k in _sett.__dict__.keys() 
                          if k.startswith('StoreHit') and getattr(_sett, k)]
      rayMetadata = dict(self.metadata)
      rayMetadata.update(metadata)
      rayMetadata.update(self.__dict__)
      #print('full metadata:', rayMetadata)
      #print('enabled keys:', _enabledKeys)
      rayMetadata = {k:v for k,v in rayMetadata.items() 
                                    if k.lower() in _enabledKeys }
      #print('filtered metadata:', rayMetadata)

    else:
      # setup defaults if no settings object exists
      if maxRayLength is None:
        maxRayLength = 100 * self.lightSource.MaxRayLengthScale
      if maxIntersections is None:
        maxIntersections = 10 * self.lightSource.MaxIntersectionsScale
      
      # default to empty metadata
      rayMetadata = {}

    # counters for total ray intersections and index of active optical 
    # element in sequential mode
    sequenceIndex = 0
    numIntersections = 0

    # variables to store current state during intersection finder loop
    prevPoint, currentPoint = self.initPoint, self.initPoint
    currentDirection = self.initDirection
    prevMedium, currentMedium = None, None
    prevPower, currentPower = self.initPower, self.initPower
    colorChange = None

    # trace loop
    while True:
      # this loop may run for quite some time, keep GUI responsive by handling events
      keepGuiResponsiveAndRaiseIfSimulationDone()
      
      # stop tracing if maxIntersections limit is reached
      if numIntersections >= maxIntersections:
        break
      numIntersections += 1

      # find next intersection
      intersect = self.findNearestIntersection(currentPoint, currentDirection, 
                                               maxRayLength=maxRayLength, 
                                               sequenceIndex=sequenceIndex)
      if intersect is None:
        # if no intersection is found yield segment with maxLength and exit loop
        yield ((currentPoint, currentPoint + currentDirection/currentDirection.Length*maxRayLength), 
                currentPower, currentMedium, colorChange)
        break
      obj, face, point = intersect

      # update current state
      prevPoint, prevPower, prevMedium = currentPoint, currentPower, currentMedium
      currentPoint = point

      # add yield latest segment
      yield (prevPoint, currentPoint), prevPower, prevMedium, colorChange

      # update ray power if medium is absorptive
      if hasattr(prevMedium, 'AbsorptionLength'):
        _absLength = float(prevMedium.AbsorptionLength)
        if _absLength == 0:
          currentPower = 0
        elif isfinite(_absLength):
          currentPower = exp( -sqrt(sum( (prevPoint-currentPoint)**2 ))/_absLength )

      # calculate normal and whether ray is facing the object from the outside or the inside
      normal, isEntering = self.getNormal(face, prevPoint, currentPoint)

      # run onHit handler of object that caused intersection
      obj.Proxy.onRayHit(source=self.lightSource, obj=obj, 
                         point=currentPoint, direction=currentDirection, 
                         power=currentPower, isEntering=isEntering, 
                         metadata=rayMetadata, store=store)

      # set colorChange to value requested by the hit object
      if obj.ViewObject is not None and obj.ViewObject.Weight != 0:
        colorChange = (obj.ViewObject.Weight, obj.ViewObject.Color)
      else:
        colorChange = None

      # hit mirror: direction is mirrored at normal, 
      # medium is unchanged, power is altered according to reflectivity
      if obj.OpticalType == 'Mirror':
        currentDirection = self.mirror(currentDirection, normal)
        currentPower *= obj.Reflectivity

        # mirrors have exactly one reflection in sequential mode
        sequenceIndex += 1
           
      # hit lens: direction is altered according to Snells law,
      # medium is changed depending on whether left or entered the lens
      elif obj.OpticalType == 'Lens':
        # ray enters lens
        if isEntering:
          if currentMedium is not None:
            raise ValueError('ray entered lens while already being inside a medium, '
                             'get rid of any overlapping lenses/transmission gratings '
                             'in your project')
          n1 = 1
          currentMedium = obj
          n2 = currentMedium.RefractiveIndex

        # ray exits lens (or may suffer total reflection)
        else:
          if currentMedium is None:
            raise ValueError('ray exited lens without having entered before, '
                             'get rid of any overlapping lenses/transmission gratings '
                             'in your project')
          n1 = currentMedium.RefractiveIndex
          n2 = 1

        # update ray direction according to Snell's law
        currentDirection, isTotalReflection = self.snellsLaw(
                                      currentDirection/currentDirection.Length,
                                      n1, n2, normal)

        # if ray traveled in exit direction and not total reflection occurred,
        # set current medium to vacuum
        if not isEntering and not isTotalReflection:
          currentMedium = None

          # increment sequence index only after ray left lens 
          # (without internal reflection)
          sequenceIndex += 1

      # hit grating: direction is altered according to selected diffraction order,
      # medium never changes, entering a grating applies diffraction, leaving the
      # grating does not do anything
      elif obj.OpticalType == 'Grating':

        if obj.GratingType == 'Reflection':

          if isEntering:
            n = currentMedium.RefractiveIndex if currentMedium else 1
            currentDirection = self.lineGrating(
                                      currentDirection/currentDirection.Length,
                                      n, n, normal, obj)
            sequenceIndex += 1

          # do nothing if ray is leaving reflection grating (should never happen)
          else:
            pass
        
        elif obj.GratingType == 'Transmission':

          if isEntering:
            if currentMedium is not None:
              raise ValueError('ray entered grating while already being inside a medium, '
                               'get rid of any overlapping lenses/transmission gratings '
                               'in your project')
            n1 = 1
            currentMedium = obj
            n2 = currentMedium.RefractiveIndex
            currentDirection = self.lineGrating(
                                      currentDirection/currentDirection.Length,
                                      n1, n2, normal, obj)

          # apply lens-like refraction if ray is leaving transmission grating
          else:
            if currentMedium is None:
              raise ValueError('ray exited grating without having entered before, '
                               'get rid of any overlapping lenses/transmission gratings '
                               'in your project')
            n1 = currentMedium.RefractiveIndex
            n2 = 1

            # update ray direction according to Snell's law
            currentDirection, isTotalReflection = self.snellsLaw(
                                          currentDirection/currentDirection.Length,
                                          n1, n2, normal)

            # if ray traveled in exit direction and not total reflection occurred,
            # set current medium to vacuum
            if not isTotalReflection:
              currentMedium = None

              # increment sequence index only after ray left grating
              # (without internal reflection)
              sequenceIndex += 1

        else:
          raise ValueError(f'invalid grating type: {obj.GratingType=}')

      # hit absorber: intensity is changed / ray is ended if intensity below min
      elif obj.OpticalType == 'Absorber':
        currentPower = 0
        sequenceIndex += 1

      # hit vacuum: do nothing at all to direction / intensity
      elif obj.OpticalType == 'Vacuum':
        sequenceIndex += 1
            
      # end if beam died
      if currentPower < powerTol:
        break
  
  def _getDistTol(self, distTol):
    if distTol is None:
      distTol = 1e-2
      if settings := find.activeSimulationSettings():
        distTol = float(settings.DistanceTolerance)
    return max([distTol, 1e-6])

  def findNearestIntersection(self, start, direction, maxRayLength, distTol=None, sequenceIndex=None):
    '''
    Find the closest relevant optical object intersecting with the ray
    of given start and direction. Start and direction are expected to be
    given in global coordinates.
    '''
    distTol = self._getDistTol(distTol)

    line = Part.makeLine(start, start+direction/direction.Length*maxRayLength)
    intersects = []
    
    # loop through all relevant optical groups
    for group in find.relevantOpticalObjects(self.lightSource, sequenceIndex=sequenceIndex):

      # this loop may run for quite some time, keep GUI responsive by handling events
      keepGuiResponsiveAndRaiseIfSimulationDone()

      # only care if bounding box is closer to start point than maxRayLength and 
      # if bounding box actually intersects with the ray
      if hasattr(group, 'Shape'):
        sbb = cachedBoundBox(cachedShape(group))
        #sbb.enlarge(distTol) => for some strange reason this causes off-centered profiles in gaussian-test, keep disabled for now...
        if ( ( not isfinite(maxRayLength)
                or any([(sbb.getPoint(i)-start).Length
                                  < 10*maxRayLength 
                                        for i in range(8)]) )
            and sbb.intersect(start, direction) ):

          # loop through all faces
          for face in cachedFaces(cachedShape(group)):

            # this loop may run for quite some time, keep GUI responsive by handling events
            keepGuiResponsiveAndRaiseIfSimulationDone()

            # only care if bounding box of face intersects with ray
            fbb = cachedBoundBox(face)
            fbb.enlarge(distTol)
            if fbb.intersect(start, direction):

              # find intersection points and loop through all of them
              if intersect := line.Curve.intersect(face.Surface):
                points, _ = intersect
                for point in points:
                  
                  # this loop may run for quite some time, keep GUI responsive by handling events
                  keepGuiResponsiveAndRaiseIfSimulationDone()

                  vec = Vector(point.X, point.Y, point.Z)
                  vert = Part.Vertex(point)

                  # if found intersection point has some finite distance from 
                  # origin and lies within the target face and on the line,
                  # add to candidate list
                  if ( (vec-start).Length > distTol
                        and vert.distToShape(line)[0] < distTol
                        and vert.distToShape(face)[0] < distTol):
                    intersects.append([group, face, vec, (vec-start).Length])

    # return intersection that is closest to start (if any)
    if len(intersects):
      return sorted(intersects, key=lambda e: e[-1])[0][:-1]
  
  
  def getNormal(self, nearest_part, origin, neworigin, epsLength=1e-6):
    '''
    calculate the normal vector given, inherited from OpticsWorkbench
    '''
    dRay = neworigin - origin
    if hasattr(nearest_part, 'Surface'):
      uv = nearest_part.Surface.parameter(neworigin)
      try:
        normal = nearest_part.normalAt(uv[0], uv[1])
      except Exception:
        # try to take normal in very close vicinity, hoping it is only
        # locally illdefined
        r1, r2 = [(1 if random.random()<.5 else -1)*(.1+random.random()) for _ in range(2)]
        normal = nearest_part.normalAt(uv[0] + epsLength*r1, uv[1] + epsLength*r2)
    else:
      return Vector(0, 0, 0)
    cosangle = dRay*normal / (dRay.Length*normal.Length)
    if cosangle < 0:
      return -normal, True
    return normal, False

  def mirror(self, ray, normal):
    '''
    mirror a ray at a normal vector, inherited from OpticsWorkbench
    '''
    return -(2*normal*(ray*normal) - ray)

  def snellsLaw(self, ray, n1, n2, normal):
    '''
    apply snell's law, inherited from OpticsWorkbench
    '''
    root = 1 - n1/n2 * n1/n2 * normal.cross(ray) * normal.cross(ray)
    if root < 0: # total reflection
      return self.mirror(ray, normal), True
    return n1/n2 * normal.cross( (-normal).cross(ray)) + normal * sqrt(root), False

  def lineGrating(self, ray, n1, n2, normal, obj):  # from Ludwig 1970
    wavelength = self.initWavelength
    order = obj.GratingDiffractionOrder
    lpm = obj.GratingLinesPerMillimeter
    g_g_p_vector = obj.GratingLinesOrientation

    # get parameters
    wavelength = wavelength / 1000
    ray = ray / ray.Length
    surf_norma = normal
    surf_norma = surf_norma / surf_norma.Length  # normalize the surface normal
    # hypothetical first vector determining the orientation of the grating rules. 
    # This vector is normal to a plane that would cause the rules by intersection 
    # with the surface of the grating.
    g_g_p_vector = g_g_p_vector / g_g_p_vector.Length

    P = g_g_p_vector.cross(surf_norma)
    P = P / P.Length
    D = surf_norma.cross(P)
    D = D / D.Length
    mu = n1 / n2
    d = 1000 / lpm
    T = (order * wavelength) / (n1 * d)
    V = (mu * (ray[0] * surf_norma[0] + ray[1] * surf_norma[1] +
                ray[2] * surf_norma[2])) / surf_norma.dot(surf_norma)
    W = (mu**2 - 1 + T**2 - 2 * mu * T *
          (ray[0] * D[0] + ray[1] * D[1] + ray[2] * D[2])
          ) / surf_norma.dot(surf_norma)
    Q = ((-2 * V + ((2 * V)**2 - 4 * W)**0.5) / 2,
          (-2 * V - ((2 * V)**2 - 4 * W)**0.5) / 2)

    if obj.GratingType == 'Reflection':
      S_0 = mu * ray[0] - T * D[0] + max(Q) * surf_norma[0]
      S_1 = mu * ray[1] - T * D[1] + max(Q) * surf_norma[1]
      S_2 = mu * ray[2] - T * D[2] + max(Q) * surf_norma[2]
    elif obj.GratingType == 'Transmission':
      S_0 = mu * ray[0] - T * D[0] + min(Q) * surf_norma[0]
      S_1 = mu * ray[1] - T * D[1] + min(Q) * surf_norma[1]
      S_2 = mu * ray[2] - T * D[2] + min(Q) * surf_norma[2]
    else:
      raise ValueError(f'unexpected {obj.GratingType=}')

    return -Vector(S_0, S_1, S_2)



'''
This module implements a cache for various attributes of FreeCAD objects.
Such a cache is necessary simply because using e.g. the BoundBox property
of one and the same object multiple times creates memory leaks. Try running
the following in a FreeCAD shell:

```
for _ in range(100000):
  App.ActiveDocument.Sphere.Shape.BoundBox
```

This will cause a measurable memory usage increase. Because the ray-tracing
loop uses Shape, BoundBox and other massively, cache is required to make 
sure not to ask FreeCAD for these attributes every time they are needed. 
'''

__license__ = 'LGPL-3.0-or-later'
__copyright__ = 'Copyright 2024  W. Braun (epiray GmbH)'
__authors__ = 'P. Bredol'
__url__ = 'https://github.com/zaphB/freecad.optics_design_workbench'


import time

from .. import io

_CACHE_LUT = {}
_LAST_SIZE_REPORT = None
_CACHED_PROPERTY_CALLS = 0
_ADDED_TO_CACHE_CALLS = 0
_SIZE_REPORT_INTERVAL = 600

def cacheClear():
  global _CACHE_LUT, _LAST_SIZE_REPORT, _CACHED_PROPERTY_CALLS, _ADDED_TO_CACHE_CALLS
  io.verb(f'tracing_cache: cleared')
  _LAST_SIZE_REPORT = time.time()
  _CACHED_PROPERTY_CALLS = 0
  _ADDED_TO_CACHE_CALLS = 0
  _CACHE_LUT = {}

def cachedProperty(obj, prop):
  global _LAST_SIZE_REPORT, _CACHED_PROPERTY_CALLS, _ADDED_TO_CACHE_CALLS

  # create cache entry if not existing
  if prop not in _CACHE_LUT.keys():
    _CACHE_LUT[prop] = {}
  if obj not in _CACHE_LUT[prop].keys():
    _CACHE_LUT[prop][obj] = getattr(obj, prop)
    _ADDED_TO_CACHE_CALLS += 1
  else:
    _CACHED_PROPERTY_CALLS += 1

  # check if its time to report
  if _LAST_SIZE_REPORT is None:
    _LAST_SIZE_REPORT = time.time()
  if time.time()-_LAST_SIZE_REPORT > _SIZE_REPORT_INTERVAL:
    io.verb(f'tracing_cache: {len(_CACHE_LUT)} different properties '
            f'cached, {", ".join([str(len(e)) for e in _CACHE_LUT.values()])} '
            f'elements cached for each property respectively, '
            f'stats since last trace_cache log: {1e-6*_CACHED_PROPERTY_CALLS:.0f}M '
            f'({_CACHED_PROPERTY_CALLS/(time.time()-_LAST_SIZE_REPORT):.1e}/s) properties requested, '
            f'{1e2*(_CACHED_PROPERTY_CALLS-_ADDED_TO_CACHE_CALLS)/_CACHED_PROPERTY_CALLS:.2f}'
            f'% cached responses')
    _LAST_SIZE_REPORT = time.time()
    _CACHED_PROPERTY_CALLS = 0
    _ADDED_TO_CACHE_CALLS = 0

  # return cached result
  return _CACHE_LUT[prop][obj]

def cachedShape(obj):
  return cachedProperty(obj, 'Shape')

def cachedFaces(obj):
  return cachedProperty(obj, 'Faces')

def cachedBoundBox(obj):
  return cachedProperty(obj, 'BoundBox')
  
def cachedViewObject(obj):
  return cachedProperty(obj, 'ViewObject')
